//  
//  Author:
//       smdn <smdn@smdn.jp>
// 
//  Copyright (c) 2003-2013 smdn
// 
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

using System;
using System.Collections.Generic;
using System.Text.RegularExpressions;

using Smdn.IO.FileSystems;

namespace Smdn.Applications.BittenBaum.Graphing.Disc {
  public class Disc : Graph {
    /*
     * class members
     */
    public static Disc Create(GraphingContext context)
    {
      var parameter = context.Parameter;

      var treeDepth = context.Parameter.DirectoryTree.MaxDepth;
      var maxDepth = (0 == parameter.MaxDepth) ? treeDepth : parameter.MaxDepth;

      if (treeDepth < maxDepth)
        maxDepth = treeDepth;

      var status = context.Status;

      status.TotalEntryCount = context.Parameter.DirectoryTree.TotalEntryCount;
      status.GraphedEntryCount = 0;

      var disc = new Disc(context);
      var rootPie = new TreePie(disc.DirectoryTree, null);
      var nest = 0;

      if (context.Parameter.CalculateRangeByDriveUsage && disc.DirectoryTree.IsRoot) {
        var drive = disc.DirectoryTree.Drive;

        rootPie.Range = 1.0 - drive.TotalFreeSpace / (double)drive.TotalSize;

        disc.IsRangeDriveUsage = true;
      }
      else {
        rootPie.Range = 1.0;

        disc.IsRangeDriveUsage = false;
      }

      rootPie.SubPies = CreateSubPies(disc, disc.DirectoryTree, rootPie, maxDepth, parameter.Comparison, status, ref nest);

      disc.RootPie = rootPie;
      disc.MaxDepth = nest;

      status.GraphedEntryCount = status.TotalEntryCount;

      return disc;
    }

    private static IEnumerable<Pie> CreateSubPies(Disc disc, DirectoryTree tree, TreePie parent, int maxDepth, Comparison<DirectoryEntry> comparison, GraphingContext.GraphingStatus status, ref int nest)
    {
      // 深さが最大深度を超える場合は、解析しない
      if (maxDepth <= nest)
        return EmptyArray<Pie>.Instance;

      var n = nest;

      var pies = new List<Pie>();
      var totalSize = (double)tree.GetTotalFileSize();
      var parentRange = (totalSize == 0.0) ? 0.0 : parent.Range / totalSize;
      var argument = parent.Argument;

      // 各エントリについてパイを作成
      foreach (var entry in tree.GetSortedEnumerator(comparison)) {
        var file = entry as File;
        Pie pie = null;

        if (file == null) {
          // ツリーの場合、再帰的にパイを作成する
          var subTree = entry as DirectoryTree;
          var treePie = new TreePie(subTree, parent);

          treePie.Argument = argument;
          treePie.Range = parentRange * subTree.GetTotalFileSize();

          var nn = n + 1;

          treePie.SubPies = CreateSubPies(disc, subTree, treePie, maxDepth, comparison, status, ref nn);

          if (nest < nn)
            nest = nn;

          pie = treePie;
        }
        else {
          // ファイルの場合
          pie = new Pie(file, parent);

          pie.Argument = argument;
          pie.Range = parentRange * file.Length;
        }

        argument += pie.Range;

        pies.Add(pie);

        status.GraphedEntryCount++;
      }

      // ファイルが一つでもある場合は深さを1深く計上する
      if (n == nest && 0 < pies.Count)
        nest = n + 1;

      return pies;
    }

    /*
     * instance members
     */
    public TreePie RootPie {
      get; private set;
    }

    public int MaxDepth {
      get; private set;
    }

    public bool IsRangeDriveUsage {
      get; private set;
    }

    private Disc(GraphingContext context)
      : base(context)
    {
    }

    public override ICollection<IGraphElement> GetAllElements()
    {
      return GetAllPies(RootPie);
    }

    private ICollection<IGraphElement> GetAllPies(TreePie pie)
    {
      var pies = new List<IGraphElement>();

      foreach (var subPie in pie.SubPies) {
        pies.Add(subPie);

        var subTreePie = subPie as TreePie;

        if (subTreePie != null)
          pies.AddRange(GetAllPies(subTreePie));
      }

      return pies;
    }

    public override ICollection<IGraphElement> SearchAllElement(Predicate<IGraphElement> match)
    {
      return SearchAllPie(RootPie, match);
    }

    public override ICollection<IGraphElement> SearchElement(Predicate<IGraphElement> match)
    {
      return SearchPie(RootPie, match);
    }

    private ICollection<IGraphElement> SearchAllPie(TreePie pie, Predicate<IGraphElement> match)
    {
      var matched = new List<IGraphElement>();

      foreach (var subPie in pie.SubPies) {
        if (match == null || match(subPie))
          matched.Add(subPie);

        var subTreePie = subPie as TreePie;

        if (subTreePie != null)
          matched.AddRange(SearchAllPie(subTreePie, match));
      }

      return matched;
    }

    private ICollection<IGraphElement> SearchPie(TreePie pie, Predicate<IGraphElement> match)
    {
      var matchedPies = new List<IGraphElement>();

      foreach (var subPie in pie.SubPies) {
        if (match == null || match(subPie))
          matchedPies.Add(subPie);
      }

      return matchedPies;
    }
  }
}
