//  
//  Author:
//       smdn <smdn@smdn.jp>
// 
//  Copyright (c) 2003-2013 smdn
// 
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;

using Smdn.Collections;
using Smdn.Formats;
using Smdn.Imaging;
using Smdn.IO.FileSystems;

namespace Smdn.Applications.BittenBaum.Rendering {
  using Graphing;
  using Graphing.Disc;

  partial class Renderer {
    /// <summary>
    /// Disc用のRederer
    /// </summary>
    private class DiscRenderer : IDisposable {
      private Disc disc = null;
      private RenderingContext.RenderingParameter parameter = null;

      // パイを描画する場合の、弧の最小角度
      private const float minPieSweepAngle = 2.0E-43f; // 1.611493E-43 <= sweepAngle

      // テキストを描画する場合の、弧の最小の長さ
      private const double minTextArcLength = 16.0;

      // エントリを描画する場合の、最小の大きさ
      private const double minEntryRange = 0.001;

      private readonly bool enableGradation;

      // 寸法
      private int maxDepth;
      private SizeF imageSize;
      private PointF discCenter;
      private float discOuterDiameter;
      private float discInnerDiameter;
      private float pieWidth;
      private RectangleF discBounds;
      private double pieAlphaDivider;

      // パイの枠線を描画するときのペン
      private Pen penPieBorder = null;

      // 背景を塗りつぶすときのブラシ
      private SolidBrush brushBackground = null;

      // テキストを描画するときのブラシ
      private SolidBrush brushText = null;

      // エントリ名を描画するときのフォント・フォーマット
      private Font fontName = null;
      private StringFormat stringFormatName = null;

      // 統計を描画するときのフォント
      private Font fontStatistics = null;

      // 空き領域の色
      private static readonly Color colorFreeSpace         = Color.FromArgb(96, 0xb0, 0xb0, 0xb0);
      private static readonly Color colorFreeSpaceSurround = Color.FromArgb(96, 0xff, 0xff, 0xff);

      public DiscRenderer(RenderingContext.RenderingParameter parameter)
      {
        var disc = parameter.Graph as Disc;

        if (disc == null)
          throw new InvalidOperationException("Graph is not Disc");

        this.parameter = parameter;
        this.disc = disc;

        // XXX: PathGradientBrush of libgdiplus is kinda buggy...
        enableGradation = (!Runtime.IsRunningOnMono && parameter.EnableGradation);

        CalculateMeasure();
      }

      void IDisposable.Dispose()
      {
        if (fontName != null) {
          fontName.Dispose();
          fontName = null;
        }

        if (stringFormatName != null) {
          stringFormatName.Dispose();
          stringFormatName = null;
        }

        if (fontStatistics != null) {
          fontStatistics.Dispose();
          fontStatistics = null;
        }

        if (penPieBorder != null) {
          penPieBorder.Dispose();
          penPieBorder = null;
        }

        if (brushBackground != null) {
          brushBackground.Dispose();
          brushBackground = null;
        }

        if (brushText != null) {
          brushText.Dispose();
          brushText = null;
        }
      }

      /// <summary>
      /// 描画するためのペン・ブラシを準備する
      /// </summary>
      private void PrepareDrawingUtencils(bool prepareStatisticsFonts)
      {
        if (penPieBorder == null)
          penPieBorder = new Pen(Color.FromArgb(0x60, parameter.BackgroundColor), 1.0f);

        if (brushBackground == null)
          brushBackground = new SolidBrush(parameter.BackgroundColor);

        if (brushText == null)
          brushText = new SolidBrush(parameter.TextColor);

        if (prepareStatisticsFonts) {
          if (fontName == null && parameter.DrawNames) {
            fontName = new Font(SystemFonts.DefaultFont.FontFamily, (float)parameter.FontSize, FontStyle.Regular);

            stringFormatName = new StringFormat(StringFormat.GenericDefault);
            stringFormatName.Alignment = StringAlignment.Center;
            stringFormatName.LineAlignment = StringAlignment.Center;
            stringFormatName.Trimming = StringTrimming.EllipsisCharacter;
            stringFormatName.FormatFlags = StringFormatFlags.NoWrap;
          }

          if (fontStatistics == null && parameter.DrawStatistics)
            fontStatistics = new Font(SystemFonts.DefaultFont.FontFamily, (float)parameter.FontSize, FontStyle.Regular);
        }

        if (enableGradation)
          InitializePieGradientBrushConsts();
      }

      /// <summary>
      /// ディスクの寸法を計算する
      /// </summary>
      private void CalculateMeasure()
      {
        // 描画する最大深度
        maxDepth = parameter.MaxDepth;

        if (disc.MaxDepth < maxDepth)
          maxDepth = disc.MaxDepth;
        if (maxDepth == 0)
          maxDepth = disc.MaxDepth;

        imageSize = new SizeF((float)parameter.OutputWidth - 1.0f, (float)parameter.OutputHeight - 1.0f);
        discOuterDiameter = (imageSize.Width < imageSize.Height) ? imageSize.Width : imageSize.Height; // 外径
        discInnerDiameter = discOuterDiameter * parameter.InnerDiameterPercent; // 内径
        pieWidth = ((discOuterDiameter - discInnerDiameter) / 2.0f) / (float)maxDepth; // パイの幅(半径)
        discBounds = new RectangleF((imageSize.Width - discOuterDiameter) / 2.0f, (imageSize.Height - discOuterDiameter) / 2.0f, discOuterDiameter, discOuterDiameter); // 描画範囲
        discCenter = new PointF(imageSize.Width / 2.0f, imageSize.Height / 2.0f); // 円心
        pieAlphaDivider = 1.0 / (maxDepth + 1);// パイを描画するときのα値の除数
      }

      /// <summary>
      /// パイの外径の領域を計算する
      /// </summary>
      private RectangleF CalculatePieOuterDiameterBounds(int depth)
      {
        var outerDiameter = discBounds.Width - pieWidth * 2.0f * (maxDepth - depth);

        return new RectangleF(discBounds.X + (discBounds.Width - outerDiameter) / 2.0f, discBounds.Y + (discBounds.Height - outerDiameter) / 2.0f, outerDiameter, outerDiameter);
      }

      /// <summary>
      /// Depth順に並べ替えたPieを取得する
      /// </summary>
      /// <remarks>
      /// <param name="minEntryRange">minEntryRange</param>で指定された値よりも小さいPieは含まない
      /// </remarks>
      private List<List<Pie>> GetDepthSortedPies(Disc disc, int maxDepth, double minEntryRange)
      {
        // 一番深いパイから描画するために、描画順にキューに入れる
        var enumeration = new Queue<TreePie>();
        var sorted = new List<List<Pie>>(maxDepth);

        for (var i = 0; i < maxDepth; i++) {
          sorted.Add(new List<Pie>());
        }

        enumeration.Enqueue(disc.RootPie);

        for (;;) {
          if (enumeration.Count == 0)
            break;

          var currentTreePie = enumeration.Dequeue();
          var currentDepth = currentTreePie.Depth;

          foreach (var pie in currentTreePie.SubPies) {
            if (minEntryRange <= pie.Range)
              sorted[currentDepth].Add(pie);

            var treePie = pie as TreePie;

            if (treePie != null && currentDepth <= maxDepth)
              enumeration.Enqueue(treePie);
          }
        }

        return sorted;
      }

      /// <summary>
      /// 描画する
      /// </summary>
      public void Render(Graphics g, RenderingContext.RenderingStatus status)
      {
        PrepareDrawingUtencils(parameter.DrawStatistics || parameter.DrawNames);

        // ディスクの外径を描画
        g.FillEllipse(brushBackground, discBounds);
        g.DrawEllipse(penPieBorder, discBounds);

        // パイを階層に分けて配列に格納する
        List<List<Pie>> pies;

        if (parameter.DrawSmallEntries)
          pies = GetDepthSortedPies(disc, maxDepth, 0.0);
        else
          // 微少なエントリを描画しない場合
          pies = GetDepthSortedPies(disc, maxDepth, minEntryRange);

        // 描画対象のパイの数を計上
        status.RenderedEntryCount = 0;
        status.TotalEntryCount = 0;

        foreach (var p in pies) {
          status.TotalEntryCount += p.Count;
        }

        // 深い階層から順に描画していく
        for (var depth = Math.Min(maxDepth, pies.Count); depth != 0; depth--) {
          // パイの外径
          var outerDiameterBounds = CalculatePieOuterDiameterBounds(depth);

          // 外径より内側を塗りつぶして一つ前の深さで描画した線を削除する
          if (depth != maxDepth)
            g.FillEllipse(brushBackground, outerDiameterBounds);

          if (parameter.ColoringMode == ColoringMode.Directory ||
              parameter.ColoringMode == ColoringMode.DirectoryBaumkuchen) {
            Pie lastParent = null;
            Color colorPie = Color.Empty;

            foreach (var pie in pies[depth - 1/* one based to zero based*/]) {
              status.RenderedEntryCount++;

              if (depth == 1 || pie.Parent != lastParent) {
                // 描画する色を親のパイにあわせる
                var ancestorPie = pie.GetParent(1);

                colorPie = ColorModel.ToRgb(CalculateDirectoryPieColor(ancestorPie, depth));

                lastParent = pie.Parent;
              }

              // パイを描画する
              DrawPie(pie, false, parameter.DrawNames, g, colorPie, null, outerDiameterBounds);
            }
          }
          else if (parameter.ColoringMode == ColoringMode.Volume) {
            foreach (var pie in pies[depth - 1/* one based to zero based*/]) {
              status.RenderedEntryCount++;

              var colorPie = ColorModel.ToRgb(CalculateVolumePieColor(pie, depth, parameter.ColoringHue));

              // パイを描画する
              DrawPie(pie, false, parameter.DrawNames, g, colorPie, null, outerDiameterBounds);
            }
          }
        }

        // 空き領域を描画する
        if (disc.IsRangeDriveUsage) {
          var argument = (float)disc.RootPie.Range * 360.0f - 90.0f;
          var sweep = (float)((1.0 - disc.RootPie.Range) * 360.0f);

          if (!enableGradation) {
            using (var brushPie = new SolidBrush(colorFreeSpace)) {
              g.FillPie(brushPie,
                        discBounds.X,
                        discBounds.Y,
                        discBounds.Width,
                        discBounds.Height,
                        argument,
                        sweep);
            }
          }
          else if (sweep < minPieSweepAngle) {
            // do nothing; PathGradientBrush throws OutOfMemoryException if sweepAngle < minPieSweepAngle
          }
          else {
            using (var gp = new GraphicsPath()) {
              gp.AddPie(discBounds.X, discBounds.Y, discBounds.Width, discBounds.Height, argument, sweep);

              using (var pgb = new PathGradientBrush(gp)) {
                pgb.CenterPoint = discCenter;
                pgb.CenterColor = colorFreeSpace;
                pgb.SurroundColors = new[] {colorPieGradientSurround};

                var blend = new Blend();

                blend.Positions = new[] {0.0f, 1.0f};
                blend.Factors = new[] {1.0f, 0.25f};

                pgb.Blend = blend;

                g.FillPath(pgb, gp);
              }
            }
          }
        }

        // 中心を塗りつぶす
        var discCenterBounds = new RectangleF(discCenter.X - discInnerDiameter / 2.0f, discCenter.Y - discInnerDiameter / 2.0f, discInnerDiameter, discInnerDiameter);

        g.FillEllipse(brushBackground, discCenterBounds);
        g.DrawEllipse(penPieBorder, discCenterBounds);

        // 統計情報を描画する
        if (parameter.DrawStatistics)
          DrawStatistics(g);
      }

      /// <summary>
      /// パイの色をHSI表色系で返す
      /// </summary>
      private HsvColor CalculateDirectoryPieColor(Pie pie, int pieDepth)
      {
        // 深さに応じてα値を設定
        return new HsvColor((int)Math.Truncate(360.0 * (pie.Argument + pie.Range * 0.5)),
                            160,
                            240,
                            (byte)(31.0 + 224.0 * (maxDepth - pieDepth + 1) * pieAlphaDivider));
      }

      private HsvColor CalculateVolumePieColor(Pie pie, int pieDepth, int hue)
      {
        const int volumeEmphasizeLevel = 3;
        const double rightAngle = Math.PI * 0.5;

        var volume = pie.Range / (pie.Parent == null ? 1.0 : pie.Parent.Range);

        // 値をsinで強調
        for (var emphasize = 0; emphasize < volumeEmphasizeLevel; emphasize++) {
          volume = Math.Sin(volume * rightAngle);
        }

        // 深さに応じてα値を設定
        return new HsvColor(hue,
                            (byte)(255 * volume),
                            255,
                            (byte)(31.0 + 224.0 * (maxDepth - pieDepth + 1) * pieAlphaDivider));
      }

      private Blend pieGradientBrushBlend = null;
      private Color[] pieGradientBrushSurroundColors;

      private const int pieBaumkuchenLayerCount = 1;
      private static readonly Color colorPieGradientSurround = Color.White;
      private static readonly Color colorBaumkuchenBaseColor = Color.FromArgb(251, 231, 159);
      private static readonly Color colorBaumkuchenLayerColor = Color.FromArgb(183, 102, 13);
      private static readonly Color colorBaumkuchenHighlightColor = Color.FromArgb(156, 222, 255);

      private void InitializePieGradientBrushConsts()
      {
        pieGradientBrushBlend = new Blend();

        if (parameter.ColoringMode == ColoringMode.DirectoryBaumkuchen) {
          var factorCount = 3 + pieBaumkuchenLayerCount * 4;

          pieGradientBrushBlend.Positions = new float[factorCount];
          pieGradientBrushBlend.Factors   = new float[factorCount];

          pieGradientBrushBlend.Positions[0] = 0.0f;
          pieGradientBrushBlend.Factors[0] = 0.0f; // layer color

          pieGradientBrushBlend.Positions[factorCount - 1] = 1.0f;
          pieGradientBrushBlend.Factors[factorCount - 1] = 1.0f; // base color

          pieGradientBrushSurroundColors = new Color[] {colorBaumkuchenLayerColor};
        }
        else /* if ColoringMode.Directory */ {
          pieGradientBrushBlend.Positions = new[] {0.0f, 1.0f};
          pieGradientBrushBlend.Factors = new[] {1.0f, 0.25f};

          pieGradientBrushSurroundColors = new Color[] {colorPieGradientSurround};
        }
      }

      /// <summary>
      /// パイを描画する
      /// </summary>
      private void DrawPie(Pie pie, bool drawInside, bool drawNames, Graphics g, Color colorPie, Pen penHighlightedBorder, RectangleF pieOuterDiameterBounds)
      {
        // 描画する弧の長さ
        var arcLength = Math.PI * pieOuterDiameterBounds.Width * pie.Range;

        // パイを描画
        var argument = (float)pie.Argument * 360.0f - 90.0f;
        var sweep = (float)(pie.Range * 360.0f);

        if (pieGradientBrushBlend == null) {
          using (var brushPie = new SolidBrush(colorPie)) {
            g.FillPie(brushPie, pieOuterDiameterBounds.X, pieOuterDiameterBounds.Y, pieOuterDiameterBounds.Width, pieOuterDiameterBounds.Height, argument, sweep);
          }
        }
        else if (sweep < minPieSweepAngle) {
          // do nothing; PathGradientBrush throws OutOfMemoryException if sweepAngle < minPieSweepAngle
        }
        else {
          if (parameter.ColoringMode == ColoringMode.DirectoryBaumkuchen) {
            colorPie = colorBaumkuchenBaseColor; // replace

            var innerDiameter = (pieOuterDiameterBounds.Width - pieWidth * 2.0f) / pieOuterDiameterBounds.Width;
            var layerDiameter = ((1.0f - innerDiameter) / pieBaumkuchenLayerCount) / 4.0f;

            var blendPositions = pieGradientBrushBlend.Positions;
            var blendFactors = pieGradientBrushBlend.Factors;

            for (var layer = 0; layer < pieBaumkuchenLayerCount; layer++) {
              var index = 1 + layer * 4;

              blendPositions[index] = (1.0f - innerDiameter) * (layer / (float)pieBaumkuchenLayerCount);
              blendFactors[index]   = 1.0f; // base color

              blendPositions[index + 1] = blendPositions[index] + layerDiameter / 8.0f;
              blendFactors[index + 1] = 0.0f; // layer color

              blendPositions[index + 2] = blendPositions[index + 1] + layerDiameter / 2.0f;
              blendFactors[index + 2] = 0.5f; // layer color (gradient)

              blendPositions[index + 3] = blendPositions[index + 2] + layerDiameter / 8.0f;
              blendFactors[index + 3] = 0.75f; // base color (gradient)
            }

            blendPositions[blendPositions.Length - 2] = 1.0f - innerDiameter;
            blendFactors[blendFactors.Length - 2] = 1.0f; // base color
          }

          using (var gp = new GraphicsPath()) {
            gp.AddPie(pieOuterDiameterBounds.X,
                      pieOuterDiameterBounds.Y,
                      pieOuterDiameterBounds.Width,
                      pieOuterDiameterBounds.Height,
                      argument,
                      sweep);

            using (var pgb = new PathGradientBrush(gp)) {
              pgb.CenterPoint = new PointF(pieOuterDiameterBounds.Left + pieOuterDiameterBounds.Width / 2.0f,
                                           pieOuterDiameterBounds.Top + pieOuterDiameterBounds.Height / 2.0f);
              pgb.CenterColor = colorPie;
              pgb.SurroundColors = pieGradientBrushSurroundColors;
              pgb.Blend = pieGradientBrushBlend;

              g.FillPath(pgb, gp);
            }
          }
        }

        // 枠を描画
        if (penHighlightedBorder != null)
          // ハイライト用のペンで描画する
          g.DrawPie(penHighlightedBorder, pieOuterDiameterBounds, argument, sweep);
        else if (penPieBorder.Width <= arcLength)
          // 描画する弧の長さがペンの幅より大きい場合のみ枠線を描画する
          g.DrawPie(penPieBorder, pieOuterDiameterBounds, argument, sweep);

        // 内径の境界線を描画する場合
        if (drawInside) {
          var pieInnerDiameterBounds = RectangleF.FromLTRB(pieOuterDiameterBounds.Left + pieWidth,
                                                           pieOuterDiameterBounds.Top + pieWidth,
                                                           pieOuterDiameterBounds.Right - pieWidth,
                                                           pieOuterDiameterBounds.Bottom - pieWidth);

          g.DrawPie(penPieBorder, pieInnerDiameterBounds, argument, sweep);
        }

        // 弧の長さが24px以上の場合は、名称を描画
        if (drawNames && minTextArcLength < arcLength)
          DrawEntryName(pie, g, pieOuterDiameterBounds);
      }

      /// <summary>
      /// パイにエントリ名を描画する
      /// </summary>
      private void DrawEntryName(Pie pie, Graphics g, RectangleF pieOuterDiameterBounds)
      {
        var characterOffset = 0.0f;
        var radius = pieOuterDiameterBounds.Width / 2.0;
        var centerX = pieOuterDiameterBounds.X + radius;
        var centerY = pieOuterDiameterBounds.Y + radius;
        var textRadius = radius - fontName.Height * 0.5;
        var arcLength = 2.0 * Math.PI * textRadius * pie.Range;
        var fontEmSize = fontName.SizeInPoints * g.DpiX / 72.0f;

        // エントリ名
        var name = pie.Entry.Name;

        // ディレクトリの場合は末尾にセパレータを追加する
        if (pie.Entry is Directory)
          name += System.IO.Path.DirectorySeparatorChar;

        // 描画するテキストの全長から初期回転角を決める
        double initialRotationRadian;

        using (var textPath = new GraphicsPath()) {
          textPath.AddString(name,
                             fontName.FontFamily,
                             (int)fontName.Style,
                             fontEmSize,
                             Point.Empty,
                             StringFormat.GenericDefault);

          var textLength = textPath.GetBounds().Width;

          if (arcLength < textLength) {
            // 左詰め
            initialRotationRadian = pie.Argument * 2.0 * Math.PI;
          }
          else {
            // センタリング
            initialRotationRadian = (pie.Argument + pie.Range / 2.0) * 2.0 * Math.PI;
            initialRotationRadian -= (textLength / 2.0) / textRadius;
          }
        }

        for (var i = 0; i < name.Length; i++) {
          var character = name.Substring(i, 1);

          using (var characterPath = new GraphicsPath()) {
            characterPath.AddString(character,
                                    fontName.FontFamily,
                                    (int)fontName.Style,
                                    fontEmSize,
                                    Point.Empty,
                                    StringFormat.GenericTypographic);

            var characterBounds = characterPath.GetBounds();

            if (arcLength < characterOffset + characterBounds.Width)
              // 弧の範囲を超えるので描画を中断
              break;

            var rotationRadian = initialRotationRadian + (characterOffset / textRadius);
            var characterOffsetRadian = rotationRadian - Math.PI / 2.0;
            var characterRotationRadian = rotationRadian + (characterBounds.Width / 2.0) / textRadius;
            var transform = new Matrix();

            transform.Translate((float)(centerX + textRadius * Math.Cos(characterOffsetRadian)),
                                (float)(centerY + textRadius * Math.Sin(characterOffsetRadian)));
            transform.Rotate((float)(180.0 * characterRotationRadian / Math.PI));

#if true
            characterPath.Transform(transform);

            g.FillPath(brushText, characterPath);
#else
            g.Transform = transform;
            g.DrawString(character, fontName, brushText, RectangleF.Empty, StringFormat.GenericTypographic);
#endif

            characterOffset += characterBounds.Width;
          } // using characterPath
        } // foreach character
      }

#if false
      private void DrawEntryName(Pie pie, Graphics g, RectangleF pieOuterDiameterBounds)
      {
        Matrix prevTransform = null;

        try {
          prevTransform = g.Transform;

          // パイの描画角にあわせて回転
          var rotator = new Matrix();

          rotator.RotateAt((float)(pie.Argument + pie.Range * 0.5) * 360.0f, discCenter);

          g.Transform = rotator;

          // 描画する弧の長さ
          var arcLength = (float)(Math.PI * pieOuterDiameterBounds.Width * pie.Range);

          // テキストの描画領域
          var textBounds = new RectangleF(discCenter.X - arcLength / 2.0f,
                                          pieOuterDiameterBounds.Top,
                                          arcLength,
                                          pieWidth);

          // エントリ名
          var name = pie.Entry.Name;

          // ディレクトリの場合は末尾にセパレータを追加する
          if (pie.Entry is Directory)
            name += System.IO.Path.DirectorySeparatorChar;

          g.DrawString(name, fontName, brushText, textBounds, stringFormatName);
        }
        finally {
          if (prevTransform != null)
            g.Transform = prevTransform;
        }
      }
#endif

      /// <summary>
      /// ディスクの中心にルートディレクトリの統計情報を描画する
      /// </summary>
      private void DrawStatistics(Graphics g)
      {
        var tree = disc.DirectoryTree;
        var text = string.Format(SIPrefixFormat.InvaliantInfo, "{0}\n\n{1:b2}B, {1:N0} Bytes\n{2:N0} entries ({3:N0} files, {4:N0} directories)",
                                 tree.FullName,
                                 tree.GetTotalFileSize(),
                                 tree.TotalEntryCount,
                                 tree.TotalFileCount,
                                 tree.TotalSubTreeCount);

        using (var format = new StringFormat(StringFormat.GenericDefault)) {
          format.Alignment = StringAlignment.Center;
          format.LineAlignment = StringAlignment.Center;
          format.Trimming = StringTrimming.None;

          // 内径に内接する四角形
          var discInnerRadius = discInnerDiameter / 2.0f;
          var discInsideBounds = new RectangleF(discCenter.X - discInnerRadius * 0.707f,
                                                discCenter.Y - discInnerRadius * 0.707f,
                                                discInnerRadius * 1.414f,
                                                discInnerRadius * 1.414f);

          g.DrawString(text, fontStatistics, brushText, discInsideBounds, format);
        }
      }

      /// <summary>
      /// Disc上の任意の位置に描画されるPieを取得する
      /// </summary>
      public Pie GetPieAt(PointF point)
      {
        // 点の位置から深さと範囲を算出する
        var dx = point.X - discCenter.X;
        var dy = point.Y - discCenter.Y;
        var diameter = 2.0 * Math.Sqrt(dx * dx + dy * dy);

        uint depth = 0;

        if (diameter < (double)discInnerDiameter)
          // 中心からの距離の2倍が内径に満たない場合は、ディスクの内側
          return null;
        else if ((double)discOuterDiameter < diameter)
          // 中心からの距離の2倍が外径を超える場合は、ディスクの外側
          return null;
        else
          depth = (uint)((diameter - discInnerDiameter) / (pieWidth * 2.0)) + 1;

        var range = 0.0;

        // Math.Atan2 = -π < θ ≦ π
        var arg = Math.Atan2(dy, dx) / (Math.PI * 2.0);

        if (arg < 0.0)
          arg += 1.0;

        range = arg - 0.75;

        if (range < 0.0)
          range = range + 1.0;

        if (disc.IsRangeDriveUsage && (disc.RootPie.Range < range && range <= 1.0))
          return Pie.FreeSpace;

        // depth、rangeにマッチするパイを探す
        var enumeration = new Queue<TreePie>();
        var sorted = new List<List<Pie>>();

        sorted.Add(new List<Pie>());

        for (var current = disc.RootPie; ; ) {
          foreach (var pie in current.SubPies) {
            if (pie.Depth < depth) {
              if (sorted.Count == pie.Depth)
                sorted.Add(new List<Pie>());

              sorted[(int)pie.Depth].Add(pie);
            }
            else if (depth == pie.Depth) {
              arg = range - pie.Argument;

              if (0.0 <= arg && arg < pie.Range)
                return pie;
            }
            else {
              return null;
            }

            var treePie = pie as TreePie;

            if (treePie != null)
              enumeration.Enqueue(treePie);
          }

          if (enumeration.Count == 0)
            break;

          current = enumeration.Dequeue();
        }

        return null;
      }

      /// <summary>
      /// Disc上の任意のPieを描画したBitmapを取得する
      /// </summary>
      public PartialGraphImage GetPieImage(IEnumerable<IGraphElement> elements, bool highlighted)
      {
        // 最深の深さを取得すると同時に、深さ別に分ける
        var deepest = 0;
        var rangedPies = new Dictionary<int, List<Pie>>();

        foreach (var pie in elements.Cast<Pie>()) {
          if (deepest < pie.Depth)
            deepest = pie.Depth;

          if (!rangedPies.ContainsKey(pie.Depth))
            rangedPies[pie.Depth] = new List<Pie>();

          rangedPies[pie.Depth].Add(pie);
        }

        // 寸法と最深を元に画像のサイズを決める
        var partialGraphSize = (int)(discInnerDiameter + (deepest - disc.RootPie.Depth) * pieWidth * 2.0f);

        // 戻り値のインスタンスを作成(大きさは余分にとる)
        var partialGraph = new PartialGraphImage(elements, new Bitmap(partialGraphSize + 1, partialGraphSize + 1));

        // 描画するときのオフセットを求める
        partialGraph.ImageOffsetX = (int)(discBounds.X + (discBounds.Width - (float)partialGraphSize) / 2.0f);
        partialGraph.ImageOffsetY = (int)(discBounds.Y + (discBounds.Height - (float)partialGraphSize) / 2.0f);

        // 画像全体に描画するため、領域の中心をあわせる
        discCenter = new PointF(partialGraphSize / 2.0f, partialGraphSize / 2.0f);

        discBounds.X = discCenter.X - discBounds.Width / 2.0f;
        discBounds.Y = discCenter.Y - discBounds.Height / 2.0f;

        // 枠線を描画するためのペンを準備する
        PrepareDrawingUtencils(false);

        using (var g = Graphics.FromImage(partialGraph.Image)) {
          g.Clear(Color.Transparent);

          g.SmoothingMode = parameter.SmoothingMode;
          g.InterpolationMode = parameter.InterpolationMode;
          g.TextRenderingHint = parameter.TextRenderingHint;

          // 深い順に描画する
          for (var depth = deepest; disc.RootPie.Depth != depth; depth--) {
            var outerDiameterBounds = CalculatePieOuterDiameterBounds(depth);

            // 内側を塗りつぶす
            g.FillEllipse(brushBackground, outerDiameterBounds);

            // 描画すべきパイがない場合は次の深さへ
            List<Pie> rangedPie;

            if (!rangedPies.TryGetValue(depth, out rangedPie))
              continue;

            foreach (var pie in rangedPie) {
              var parentPie = pie.GetParent(1);

              HsvColor hsvFill;
              Color colorBorder;

              if (parameter.ColoringMode == ColoringMode.Directory || 
                  parameter.ColoringMode == ColoringMode.DirectoryBaumkuchen) {
                hsvFill = CalculateDirectoryPieColor(parentPie, parentPie.Depth);

                if (parameter.ColoringMode == ColoringMode.DirectoryBaumkuchen) {
                  colorBorder = colorBaumkuchenHighlightColor;
                }
                else {
                  var hsvBorder = hsvFill;

                  hsvBorder.S = 255;
                  hsvBorder.V = 192;

                  colorBorder = ColorModel.ToRgb(hsvBorder);
                }
              }
              else if (parameter.ColoringMode == ColoringMode.Volume) {
                hsvFill = CalculateVolumePieColor(parentPie, parentPie.Depth, parameter.ColoringHue);

                if (highlighted)
                  hsvFill.H += 30;

                colorBorder = ColorModel.ToRgb(hsvFill);
              }
              else {
                hsvFill = HsvColor.Empty;
                colorBorder = Color.Empty;
              }

              if (highlighted) {
                using (var penHighlightedBorder = new Pen(colorBorder, 4.0f)) {
                  DrawPie(pie, true, parameter.DrawNames, g, ColorModel.ToRgb(hsvFill), penHighlightedBorder, outerDiameterBounds);
                }
              }
              else {
                DrawPie(pie, true, parameter.DrawNames, g, ColorModel.ToRgb(hsvFill), null, outerDiameterBounds);
              }
            }
          }

          // 中心を塗りつぶす
          var discCenterBounds = new RectangleF(discCenter.X - discInnerDiameter / 2.0f,
                                                discCenter.Y - discInnerDiameter / 2.0f,
                                                discInnerDiameter,
                                                discInnerDiameter);

          g.FillEllipse(brushBackground, discCenterBounds);
        }

        // 塗りつぶした箇所を透過させる
        partialGraph.Image.MakeTransparent(parameter.BackgroundColor);

        return partialGraph;
      }
    } // class DiscRenderer
  } // class Renderer
}
