//  
//  Author:
//       smdn <smdn@smdn.jp>
// 
//  Copyright (c) 2003-2013 smdn
// 
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.Drawing.Text;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;
using System.Windows.Forms;

using Smdn.Formats;
using Smdn.IO.FileSystems;

namespace Smdn.Applications.BittenBaum.UI {
  using Analysis;
  using Formatting;
  using Graphing;
  using Rendering;

  using IO = System.IO;

  public partial class MainWindow : Form {
    private enum SaveResultFormatType {
      Graph,
      Xml,
      Csv,
    }

    private const int visualizationHistoryMax = 5;

    private Viewer viewer = null;
    private Tabulator tabulator = null;

    /// <summary>
    /// コンストラクタ
    /// </summary>
    public MainWindow()
    {
      InitializeComponent();

      Initialize();
    }

    private void Initialize()
    {
      viewer = new Viewer(this.pictureBoxViewer);
      tabulator = new Tabulator(this.listViewTabulator, this.textBoxTabulatorTitle);

      // ハンドラの割り当て
      InitializeHandler();

      // アイコンの割り当て
      toolStripMenuItemVisualize_Directory.Image = IconTheme.GetIcon(IconType.FolderSelection);
      toolStripButtonVisualizeDirectory.Image = IconTheme.GetIcon(IconType.FolderSelection);

      SetTitleText(null);

      this.StartPosition = FormStartPosition.CenterScreen;

      // スタートアップ操作用のパネルを中央に配置する
      panelStartUpOperation.Left = pictureBoxViewer.Left + (pictureBoxViewer.Width - panelStartUpOperation.Width) / 2;
      panelStartUpOperation.Top = pictureBoxViewer.Top + (pictureBoxViewer.Height - panelStartUpOperation.Height) / 2;

      // 起動直後は無効にしておく
      DisableSearchAndRefresh();

      DisableAbortVisualization();

      toolStripButtonAscendDirectory.Enabled = false;

      // 起動直後は隠しておく
      toolStripProgressBarStatus.Visible = false;

      toolStripStatusLabelStatus.Text = "解析するフォルダまたはドライブを選択してください";

      toolStripSplitButtonForwardHistory.Enabled = false;
      toolStripSplitButtonBackHistory.Enabled = false;
    }

    #region "visualization"

    private Thread visualizingThread = null;
    private ManualResetEvent visualizingEvent = new ManualResetEvent(false);
    private System.Threading.Timer visualizingTimer = null;

    private Directory SelectVisualizeDirectory()
    {
      using (var dialog = new FolderBrowserDialog()) {
        dialog.Description = "解析するフォルダを選択してください。";
        dialog.RootFolder = Environment.SpecialFolder.Desktop;
        dialog.ShowNewFolderButton = false;

        if (DialogResult.OK == dialog.ShowDialog(this)) {
          if (string.IsNullOrEmpty(dialog.SelectedPath))
            return null;

          try {
            return FileSystem.GetDirectory(dialog.SelectedPath);
          }
          catch (NotSupportedException) {
            MessageBox.Show(this,
                            string.Format("フォルダ '{0}' は解析できません。", dialog.SelectedPath),
                            Application.ProductName,
                            MessageBoxButtons.OK,
                            MessageBoxIcon.Exclamation);

            return null;
          }
        }
        else {
          return null;
        }
      }
    }

    private void Visualize(Drive drive)
    {
      switch (drive.Type) {
        case System.IO.DriveType.Fixed:
        case System.IO.DriveType.Ram:
          break;

        default:
          var result = MessageBox.Show(this,
                                       "このドライブの解析には時間がかかる可能性があります。",
                                       Application.ProductName,
                                       MessageBoxButtons.OKCancel,
                                       MessageBoxIcon.Information,
                                       MessageBoxDefaultButton.Button1);

          if (result == DialogResult.Cancel)
            return; // cancel
          else
            break;
      }

      Visualize(drive.RootDirectory);
    }

    private void AbortVisualize()
    {
      if (visualizingThread != null)
        visualizingThread.Abort();
    }

    private void EnableAbortVisualization()
    {
      toolStripButtonAbortVisualize.Enabled = true;
      toolStripMenuItemVisualize_Abort.Enabled = true;
      panelStartUpOperation.Enabled = false;

      this.Cursor = Cursors.WaitCursor;
    }

    private void DisableAbortVisualization()
    {
      toolStripButtonAbortVisualize.Enabled = false;
      toolStripMenuItemVisualize_Abort.Enabled = false;
      panelStartUpOperation.Enabled = true;

      this.Cursor = Cursors.Default;
    }

    private void RefreshDirectoryTree(/*DirectoryTree directoryTree*/)
    {
      /*
      if (directoryTree == null)
        return;

      var visualizer = new Visualizer(directoryTree);
      */

      var rootDirectoryTree = viewer.Graph.DirectoryTree;

      for (;;) {
        if (rootDirectoryTree.ParentTree == null)
          break;
        else
          rootDirectoryTree = rootDirectoryTree.ParentTree;
      }

      var visualizer = new Visualizer(rootDirectoryTree);

      BeginVisualize(visualizer, delegate {
        SetVisualizer(visualizer, false);
      });
    }

    private void Visualize(Directory directory)
    {
      if (directory == null)
        return;

      var visualizer = new Visualizer(directory);

      BeginVisualize(visualizer, delegate {
        SetVisualizer(visualizer, false);
      });
    }

    private void BeginVisualize(Visualizer visualizer, Action onVisualized)
    {
      if (visualizingEvent.WaitOne(0, false)) {
        MessageBox.Show(this,
                        "他のフォルダを解析中です。",
                        Application.ProductName,
                        MessageBoxButtons.OK,
                        MessageBoxIcon.Exclamation);

        return;
      }

      var directoryTree = visualizer.AnalysisContext.Parameter.AnalysisDirectory as DirectoryTree;
      string directoryName;

      if (directoryTree == null) {
        var directory = visualizer.AnalysisContext.Parameter.AnalysisDirectory;

        directoryName = directory.FullName;

        // 既に同じディレクトリを解析している場合は、そのVisualizerを再設定する
        var v = GetVisualizedHistory(directory);

        if (v != null) {
          SetVisualizer(v, true);

          return;
        }
      }
      else {
        directoryName = directoryTree.FullName;
      }

      // 開始
      visualizingEvent.Set();

      SetStatus(0.0f, string.Format("ディレクトリ'{0}'の解析を開始します", directoryName));

      EnableAbortVisualization();

      visualizingThread = new Thread(VisualizeCore);

      visualizingThread.Start(new object[] {visualizer, onVisualized});

      // タイマーを起動して定期的に監視する
      TimerCallback visualizingState = delegate(object state) {
        if (visualizer.Status == null)
          return;

        var progress = 0.0f;
        var status = string.Empty;

        lock (visualizer.Status) {
          progress = visualizer.Status.Percentage;
          status = visualizer.Status.StatusText;
        }

        //progress = ((float)visualizer.TaskCompleted + progress) / (float)visualizer.TaskTotal;

        // ステータスを設定
        this.BeginInvoke(new SetStatusDelegate(SetStatus), progress, status);
      };

      visualizingTimer = new System.Threading.Timer(visualizingState, visualizingThread, 0, 250);
    }

    private void VisualizeCore(object parameter)
    {
      var status = string.Empty;
      var aborted = false;
      var parameters = parameter as object[];
      var visualizer = parameters[0] as Visualizer;
      var onVisualized = parameters[1] as Action;

      try {
        try {
          Config.VisualizerGraphOptionDefaults.Apply(visualizer.GraphingContext.Parameter);

          visualizer.GraphingContext.Parameter.Type = GraphType.Disc;

          var renderingParameter = visualizer.RenderingContext.Parameter;

          Config.VisualizerGraphOptionDefaults.Apply(renderingParameter);

          renderingParameter.DrawNames = false;
          renderingParameter.DrawStatistics = false;
          renderingParameter.InnerDiameterPercent = 0.25f;
          renderingParameter.OutputWidth = Math.Min(Screen.PrimaryScreen.Bounds.Width * 4 / 5, Screen.PrimaryScreen.Bounds.Height * 4 / 5);
          renderingParameter.OutputHeight = renderingParameter.OutputWidth;

          var stopwatch = Stopwatch.StartNew();

          stopwatch.Start();

          visualizer.Visualize();

          stopwatch.Stop();

          status = string.Format("{0} ({1:N2}秒)",
                                 renderingParameter.ColoringMode == ColoringMode.DirectoryBaumkuchen ? "バウムクーヘンが焼けました" : "完了しました",
                                 stopwatch.Elapsed.TotalSeconds);
        }
        catch (ThreadAbortException) {
          // 中断要求
          status = "中断しました";
          aborted = true;
          visualizer = null;
        }
        catch (Exception ex) {
          status = "エラーが発生しました: " + ex.Message;
          visualizer = null;
          Console.Error.WriteLine(ex);
        }
      }
      finally {
        this.BeginInvoke(new Action<Visualizer, bool, string, Action>(VisualizeFinished), visualizer, aborted, status, onVisualized);
      }
    }

    private void VisualizeFinished(Visualizer visualized, bool aborted, string status, Action onVisualized)
    {
      try {
        if (visualized == null) {
          if (!aborted)
            MessageBox.Show(this,
                            "解析中にエラーが発生しました。",
                            Application.ProductName,
                            MessageBoxButtons.OK,
                            MessageBoxIcon.Error);
        }
        else {
          SetStatus(1.0f, "結果を表示しています");

          onVisualized();
        }

        visualizingTimer.Dispose();

        SetFinalStatus(status);
      }
      finally {
        DisableAbortVisualization();

        visualizingThread = null;

        visualizingEvent.Reset();
      }
    }

    #endregion

    #region "visualization history"

    private List<Visualizer> visualized = new List<Visualizer>(visualizationHistoryMax + 1);

    private Visualizer GetVisualizedHistory(Directory direcotry)
    {
      foreach (var v in visualized) {
        if (Smdn.IO.PathUtils.ArePathEqual(v.AnalysisContext.Parameter.AnalysisDirectory.FullName, direcotry.FullName))
          return v;
      }

      return null;
    }

    private void AddVisualizedHistory(Visualizer visualizer)
    {
      visualized.Add(visualizer);

      EventHandler clickVisualizedHistoryHandler = delegate(object sender, EventArgs e) {
        SetVisualizer((sender as IVisualizerControl).Visualizer, true);
      };

      var button = new VisualizedHistoryToolStripButton(visualizer);

      button.Image = IconTheme.GetIcon(visualizer.AnalysisContext.Parameter.AnalysisDirectory);
      button.Text = visualizer.AnalysisContext.Parameter.AnalysisDirectory.FullName;
      button.ToolTipText = button.Text;
      button.DisplayStyle = ToolStripItemDisplayStyle.Image;
      button.Click += clickVisualizedHistoryHandler;

      toolStripMain.Items.Add(button);

      var item = new VisualizedHistoryToolStripMenuItem(visualizer);

      item.Image = button.Image;
      item.Text = button.Text;
      item.Click += clickVisualizedHistoryHandler;

      var indexToInsert = toolStripMenuItemVisualize_History.DropDownItems.IndexOf(toolStripMenuItemVisualize_History_Clear);

      toolStripMenuItemVisualize_History.DropDownItems.Insert(indexToInsert - 1, item);

      toolStripMenuItemVisualize_History_Clear.Enabled = (0 < visualized.Count);

      if (visualized.Count <= visualizationHistoryMax)
        return;

      // 追加したことにより履歴が5件を超える場合は、一番古い履歴を削除する
      visualized.RemoveAt(0);

      VisualizedHistoryToolStripButton firstButton = null;

      foreach (var b in toolStripMain.Items) {
        firstButton = b as VisualizedHistoryToolStripButton;

        if (firstButton != null)
          break;
      }

      if (firstButton != null)
        toolStripMain.Items.Remove(firstButton);

      VisualizedHistoryToolStripMenuItem firstItem = null;

      foreach (var i in toolStripMenuItemVisualize_History.DropDownItems) {
        firstItem = i as VisualizedHistoryToolStripMenuItem;

        if (firstItem != null)
          break;
      }

      if (firstItem != null && !Object.ReferenceEquals(firstItem, toolStripMenuItemVisualize_History_Separator))
        toolStripMenuItemVisualize_History.DropDownItems.Remove(firstItem);
    }

    private void ClearVisualizedHistory()
    {
      visualized.Clear();

      var buttonsToRemove = new List<ToolStripItem>();
    
      foreach (ToolStripItem b in toolStripMain.Items) {
        buttonsToRemove.Add(b);
      }

      foreach (var b in buttonsToRemove) {
        if (b is VisualizedHistoryToolStripButton)
          toolStripMain.Items.Remove(b);
      }

      var itemsToRemove = new List<ToolStripItem>();
    
      foreach (ToolStripItem i in toolStripMenuItemVisualize_History.DropDownItems) {
        if (i is VisualizedHistoryToolStripMenuItem)
          itemsToRemove.Add(i);
      }

      foreach (var i in itemsToRemove) {
        toolStripMenuItemVisualize_History.DropDownItems.Remove(i);
      }

      toolStripMenuItemVisualize_History_Clear.Enabled = false;
    }
    #endregion

    #region "viewing"

    private void EnableSearchAndRefresh()
    {
      panelSearchQuery.Enabled = true;
      textBoxSearch.Text = string.Empty;

      toolStripMenuItemVisualize_Refresh.Enabled = true;
      toolStripButtonRefresh.Enabled = true;
    }

    private void DisableSearchAndRefresh()
    {
      panelSearchQuery.Enabled = false;

      toolStripMenuItemVisualize_Refresh.Enabled = false;
      toolStripButtonRefresh.Enabled = false;
    }

    private void SetVisualizer(Visualizer visualizer, bool fromHistory)
    {
      if (panelStartUpOperation.Visible)
        // スタートアップ操作用のパネルを隠す
        panelStartUpOperation.Hide();

      // 検索ボックス関連を有効にする
      EnableSearchAndRefresh();

      if (visualizer == viewer.Visualizer)
        return;

      if (fromHistory) {
        // TODO: 並べ替え順序が変更されていたら、グラフを作成しなおしたい

        // 履歴として保持されているグラフ画像はすでに破棄されているので再描画する
        Config.VisualizerGraphOptionDefaults.Apply(visualizer.RenderingContext.Parameter);

        visualizer.Render(false);
      }

      viewer.SetVisualizer(visualizer);

      // このVisualizerが履歴に存在するかどうか調べる
      if (GetVisualizedHistory(visualizer.AnalysisContext.Parameter.AnalysisDirectory) == null)
        // 履歴として記録されていない場合は、履歴として追加する
        AddVisualizedHistory(visualizer);
    }

    private void AscendDirectory()
    {
      if (viewer.Graph != null && viewer.Graph.DirectoryTree.ParentTree != null)
        ViewDirectory(viewer.Graph.DirectoryTree.ParentTree);
    }

    private void ViewDirectory(ITreeGraphElement element)
    {
      ViewDirectory(element.DirectoryTree);
    }

    private void ViewDirectory(DirectoryTree tree)
    {
      try {
        this.Cursor = Cursors.WaitCursor;

        viewer.ViewTree(tree);
      }
      finally {
        this.Cursor = Cursors.Default;
      }

      SetStatistics(tree);
    }

    private void GraphRootChanged()
    {
      SetStatistics(viewer.Graph.DirectoryTree);

      tabulator.SetElements(viewer.Graph.GetAllElements(), viewer.Graph.DirectoryTree.FullName);

      SetTitleText(viewer.Graph.DirectoryTree.FullName);

      toolStripButtonAscendDirectory.Enabled = (viewer.Graph.DirectoryTree.ParentTree != null);
    }

    private void SelectedElementChanged()
    {
      viewer.HighlightElements(tabulator.SelectedElements);
    }

    private void ConfigureGraphOptions()
    {
      if (viewer.Graph == null) // XXX
        return;

      using (var dialog = new GraphOptionDialog()) {
        Config.VisualizerGraphOptionDefaults.Apply(dialog);

        if (dialog.ShowDialog(this, true, false) != DialogResult.OK)
          return;

        Config.VisualizerGraphOptionDefaults.Store(dialog);

        if (viewer.Visualizer != null) {
          Config.VisualizerGraphOptionDefaults.Apply(viewer.Visualizer.RenderingContext.Parameter);

          var graphParameterChanged = Config.VisualizerGraphOptionDefaults.Apply(viewer.Visualizer.GraphingContext.Parameter);

          try {
            this.Cursor = Cursors.WaitCursor;

            viewer.DiscardView(graphParameterChanged);
          }
          finally {
            this.Cursor = Cursors.Default;
          }
        }

        tabulator.CompareEntryType = dialog.CompareEntryType;
      }
    }

    private void SetTabulateMax(int max)
    {
      tabulator.TabulateMax = max;
    }

    #endregion

    #region "viewing history"

    private void ForwardHistory(int step)
    {
      viewer.ForwardHistory(step);

      SetStatistics(viewer.Graph.DirectoryTree);
    }

    private void BackHistory(int step)
    {
      viewer.BackHistory(step);

      SetStatistics(viewer.Graph.DirectoryTree);
    }

    private void HistoryChanged()
    {
      DirectoryTree[] forwardTrees, backwardTrees;

      viewer.GetHistory(out forwardTrees, out backwardTrees);

      toolStripSplitButtonForwardHistory.Enabled = (0 < forwardTrees.Length);
      toolStripSplitButtonBackHistory.Enabled = (0 < backwardTrees.Length);

      // 新しいドロップダウンリストを割り当てる
      toolStripSplitButtonForwardHistory.DropDownItems.Clear();

      foreach (var tree in forwardTrees) {
        toolStripSplitButtonForwardHistory.DropDownItems.Add(tree.Directory.FullName, null, delegate(object sender, EventArgs e) {
          ForwardHistory(toolStripSplitButtonForwardHistory.DropDownItems.IndexOf(sender as ToolStripItem) + 1);
        });
      }

      toolStripSplitButtonBackHistory.DropDownItems.Clear();

      foreach (var tree in backwardTrees) {
        toolStripSplitButtonBackHistory.DropDownItems.Add(tree.Directory.FullName, null, delegate(object sender, EventArgs e) {
          BackHistory(toolStripSplitButtonBackHistory.DropDownItems.IndexOf(sender as ToolStripItem) + 1);
        });
      }
    }

    #endregion

    #region "searching"

    private void SearchByWildcard(string wildcard)
    {
      // ワイルドカードを正規表現に置き換える
      var pattern = Regex.Escape(wildcard);

      pattern = pattern.Replace("\\*", ".*");
      pattern = pattern.Replace("\\?", ".");

      SearchAndHighlightElements("^" + pattern + "$", wildcard + " に一致するファイル・フォルダ");
    }

    private void SearchByRegex(string pattern)
    {
      SearchAndHighlightElements(pattern, pattern + " に一致するファイル・フォルダ");
    }

    private void SearchAndHighlightElements(string pattern, string patternTitle)
    {
      if (viewer.Graph == null)
        return;

      Regex regex = null;

      try {
        regex = new Regex(pattern, RegexOptions.Singleline | RegexOptions.IgnoreCase);
      }
      catch (ArgumentException ex) {
        MessageBox.Show(this,
                        "正規表現に誤りがあります。" + Environment.NewLine + ex.Message,
                        Application.ProductName,
                        MessageBoxButtons.OK,
                        MessageBoxIcon.Exclamation);

        return;
      }

      try {
        this.Cursor = Cursors.WaitCursor;

        var matched = viewer.Graph.SearchAllElement(delegate(IGraphElement element) {
          return regex.IsMatch(element.Entry.Name);
        });

        viewer.HighlightElements(matched);

        tabulator.SetElements(matched, patternTitle);

        if (matched.Count == 0)
          SetFinalStatus("該当するファイル・フォルダはありません");
        else
          SetFinalStatus(string.Format("該当するファイル・フォルダが{0}個見つかりました", matched.Count));
      }
      finally {
        this.Cursor = Cursors.Default;
      }
    }

    private void ClearSearch()
    {
      if (viewer.Graph == null)
        return;

      try {
        this.Cursor = Cursors.WaitCursor;

        viewer.ClearHighlight();

        tabulator.SetElements(viewer.Graph.GetAllElements(), viewer.Graph.DirectoryTree.FullName);
      }
      finally {
        this.Cursor = Cursors.Default;
      }
    }

    private void CopyTabulatingElementPathsToClipboard()
    {
      if (tabulator.TabulatingElements == null)
        return;

      var sb = new StringBuilder();

      foreach (var element in tabulator.TabulatingElements) {
        sb.AppendLine(element.Entry.FullName);
      }

      Clipboard.SetText(sb.ToString());
    }

    #endregion

    #region "saving"

    private void SaveResult(SaveResultFormatType type)
    {
      if (viewer.Visualizer == null)
        return;

      // 拡張子と表記
      string[][] extensions;

      switch(type) {
        case SaveResultFormatType.Graph:
          extensions = new string[][] {
            new[] {"png", "透過PNG形式"},
            new[] {"jpg", "JPEG形式"},
            new[] {"bmp", "ビットマップ形式(32bit)"},
          };
          break;

        case SaveResultFormatType.Xml:
          extensions = new string[][] {
            new[] {"xml", "XMLファイル"},
          };
          break;

        case SaveResultFormatType.Csv:
          extensions = new string[][] {
            new[] {"csv", "CSVファイル"},
          };
          break;

        default:
          return;
      }

      // フィルタ文字列を作成
      var filter = new StringBuilder();

      foreach (var extension in extensions) {
        if (filter.Length != 0)
          filter.Append("|");

        filter.AppendFormat("{1} (*.{0})|*.{0}", extension[0], extension[1]);
      }

      // デフォルトのファイル名
      var filename = Smdn.IO.PathUtils.ReplaceInvalidFileNameChars(viewer.Graph.DirectoryTree.Directory.Name, "_");

      using (var dialog = new SaveFileDialog()) {
        dialog.AddExtension = true;
        dialog.DefaultExt = extensions[0][0];
        dialog.FileName = string.Format("{0}.{1}", filename, extensions[0][0]);
        dialog.Filter = filter.ToString();
        dialog.FilterIndex = 0;
        dialog.InitialDirectory = Environment.GetFolderPath(Environment.SpecialFolder.Desktop);
        dialog.OverwritePrompt = true;
        dialog.RestoreDirectory = false;
        dialog.SupportMultiDottedExtensions = true;
        dialog.ValidateNames = true;

        // ダイアログを表示
        if (dialog.ShowDialog(this) != DialogResult.OK)
          return;

        filename = dialog.FileName;
      }

      object[] parameters;

      if (type == SaveResultFormatType.Graph) {
        var renderingContext = new RenderingContext();

        // グラフのオプションを設定
        using (var dialog = new GraphOptionDialog()) {
          Config.SaveGraphOptionDefaults.Apply(dialog);

          if (dialog.ShowDialog(this, false, true) != DialogResult.OK)
            return;

          Config.SaveGraphOptionDefaults.Store(dialog);

          var param = renderingContext.Parameter;

          Config.SaveGraphOptionDefaults.Apply(param);

          param.Graph = viewer.Visualizer.GraphingContext.Result.Graph;
        }

        parameters = new object[] {filename, renderingContext};
      }
      else {
        var formattingContext = new FormattingContext();

        formattingContext.Parameter.Graph = viewer.Visualizer.GraphingContext.Result.Graph;
        formattingContext.Parameter.FormattingElements = tabulator.TabulatingElements;

        switch (type) {
          case SaveResultFormatType.Xml:
            formattingContext.Parameter.Type = FormatType.Xml;
            break;

          case SaveResultFormatType.Csv:
            formattingContext.Parameter.Type = FormatType.Csv;
            break;
        }

        parameters = new object[] {filename, formattingContext};
      }

      // 別スレッドで保存処理を開始
      var t = new Thread(SaveResultThread);

      t.Start(parameters);

      this.Cursor = Cursors.WaitCursor;
    }

    private void SaveResultThread(object param)
    {
      var status = "not started";
      var succeeded = false;

      try {
        try {
          var parameters = param as object[];
          var filename = parameters[0] as string;

          using (var stream = new IO.FileStream(filename, IO.FileMode.Create, IO.FileAccess.Write, IO.FileShare.None)) {
            var renderingContext = parameters[1] as RenderingContext;

            if (renderingContext == null)
              SaveResultAsText(parameters[1] as FormattingContext, stream);
            else
              SaveResultAsImage(renderingContext, stream);

            status = string.Format("保存しました({0})", filename);

            succeeded = true;
          }
        }
        catch (ThreadAbortException) {
          // 中断要求
          status = "中断しました";
        }
        catch (Exception ex) {
          status = "エラーが発生しました: " + ex.Message;

          Console.Error.WriteLine(ex);
        }
      }
      finally {
        this.BeginInvoke(new SaveResultFinishedDelegate(SaveResultFinished), status, succeeded);
      }
    }

    private void SaveResultAsImage(RenderingContext context, IO.FileStream outputStream)
    {
      // Rendererで描画する
      (new Renderer(context)).Render();

      context.Result.Save(outputStream);
    }

    private void SaveResultAsText(FormattingContext context, IO.Stream outputStream)
    {
      context.Parameter.OutputStream = outputStream;

      // Formatterで所定の形式に整形する
      (new Formatter(context)).Format();
    }

    private delegate void SaveResultFinishedDelegate(string status, bool succeeded);

    private void SaveResultFinished(string status, bool succeeded) {
      try {
        if (!succeeded)
          MessageBox.Show(this,
                          "保存できませんでした。",
                          Application.ProductName,
                          MessageBoxButtons.OK,
                          MessageBoxIcon.Error);

        SetFinalStatus(status);
      }
      finally {
        this.Cursor = Cursors.Default;

        // TODO: 書き込み中に失敗した場合に残るファイルの削除
      }
    }

    #endregion

    #region "status/statistics"

    private delegate void SetStatusDelegate(float progress, string status);

    private void SetStatus(float progress, string statusText)
    {
      if (progress < 0.0f)
        progress = 0.0f;
      if (1.0f < progress)
        progress = 1.0f;
      if (float.IsNaN(progress))
        progress = 0.0f;

      if (!toolStripProgressBarStatus.Visible) {
        toolStripProgressBarStatus.Visible = true;
        toolStripProgressBarStatus.Minimum = 0;
        toolStripProgressBarStatus.Maximum = 100;
      }

      toolStripProgressBarStatus.Value = (int)(100.0f * progress);

      toolStripStatusLabelStatus.Text = statusText;
    }

    private void SetFinalStatus(string statusText)
    {
      if (toolStripProgressBarStatus.Visible)
        toolStripProgressBarStatus.Visible = false;

      toolStripStatusLabelStatus.Text = statusText;
    }

    private void SetStatistics(Drive drive)
    {
      textBoxStatistics.Text = string.Format(SIPrefixFormat.CurrentInfo,
                                             "{0} 空き領域{1}{2:F2} ({2:N0} Bytes)",
                                             drive.RootDirectory.FullName,
                                             Environment.NewLine,
                                             drive.TotalFreeSpace);
    }

    private void SetStatistics(DirectoryTree tree)
    {
      textBoxStatistics.Text = string.Format(SIPrefixFormat.CurrentInfo,
                                             "{0}{1}全エントリ数: {2:N0} (ファイル:{3:N0} フォルダ:{4:N0}) 総容量: {5:F2} ({5:N0} Bytes)",
                                             tree.FullName,
                                             Environment.NewLine,
                                             tree.TotalEntryCount,
                                             tree.TotalFileCount,
                                             tree.TotalSubTreeCount,
                                             tree.GetTotalFileSize());
    }

    private void SetStatistics(IGraphElement element)
    {
      var treeElement = element as ITreeGraphElement;

      if (treeElement == null) {
        var file = element.Entry as File;

        textBoxStatistics.Text = string.Format(SIPrefixFormat.CurrentInfo, "{0}{1}{2:F2} ({2:N0} bytes)",
                                               file.FullName,
                                               Environment.NewLine,
                                               file.Length);
      }
      else {
        SetStatistics(treeElement.DirectoryTree);
      }
    }

    private void SetTitleText(string newText)
    {
      if (string.IsNullOrEmpty(newText))
        this.Text = BittenBaum.VersionedApplicationName;
      else
        this.Text = string.Concat(newText, " - ", BittenBaum.VersionedApplicationName);
    }
#endregion

  }
}
