//  
//  Author:
//       smdn <smdn@smdn.jp>
// 
//  Copyright (c) 2003-2013 smdn
// 
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

using System;
using System.Diagnostics;
using System.Collections.Generic;
using System.Windows.Forms;

using Smdn.Collections;
using Smdn.Formats;
using Smdn.IO.FileSystems;

namespace Smdn.Applications.BittenBaum.UI {
  using Analysis;
  using Graphing;

  public class Tabulator {
    /*
     * inner types
     */
    private abstract class ElementListViewItem : ListViewItem {
      public IGraphElement Element {
        get; private set;
      }

      protected ElementListViewItem(IGraphElement element)
      {
        this.Element = element;
      }

      public static ElementListViewItem Create(IGraphElement element)
      {
        var treeElement = element as ITreeGraphElement;

        if (treeElement == null) {
#if DEBUG
          if (!(element.Entry is File))
            throw new NotImplementedException("unsupported directory entry"); // TODO
#endif
          return new FileListViewItem(element);
        }
        else {
          return new DirectoryTreeListViewItem(treeElement);
        }
      }
    }

    private class DirectoryTreeListViewItem : ElementListViewItem {
      public DirectoryTreeListViewItem(ITreeGraphElement element)
        : base(element)
      {
        this.Text = element.DirectoryTree.Name + System.IO.Path.DirectorySeparatorChar;
        this.SubItems.Add("(フォルダ)");
        this.SubItems.Add(element.Range.ToString("P2"));
        this.SubItems.Add(string.Format(SIPrefixFormat.InvaliantInfo, "{0:b}B", element.DirectoryTree.GetTotalFileSize()));
        this.SubItems.Add(element.DirectoryTree.FullName);
      }
    }

    private class FileListViewItem : ElementListViewItem {
      public FileListViewItem(IGraphElement element)
        : base(element)
      {
        var file = element.Entry as File;

        this.Text = file.Name;
        this.SubItems.Add(System.IO.Path.GetExtension(file.Name));
        this.SubItems.Add(element.Range.ToString("P2"));
        this.SubItems.Add(string.Format(SIPrefixFormat.InvaliantInfo, "{0:b}B", file.Length));
        this.SubItems.Add(file.FullName);
      }
    }

    private class EntryListViewColumn : ColumnHeader {
      public DirectoryEntryCompareMode CompareMode {
        get; private set;
      }

      public EntryListViewColumn(string text, int width, DirectoryEntryCompareMode compareMode)
        : base()
      {
        this.CompareMode = compareMode;

        base.Text = text;
        base.Width = width;
      }
    }

    /*
     * instance members
     */
    public event EventHandler<TabulatorMouseEventArgs> ItemClicked = null;
    public event EventHandler SelectedElementChanged = null;

    public ListView TabulatingListView {
      get { return tabulatingListView; }
    }

    public TextBox TabulatingTitleTextBox {
      get { return tabulatingTitleTextBox; }
    }

    public int TabulateMax {
      get { return tabulateMax; }
      set {
        if (value < 0)
          throw new ArgumentOutOfRangeException("TabulateMax");
        else if (value == tabulateMax)
          return;

        tabulateMax = value;

        if (tabulatingElements != null)
          UpdateListItem();
      }
    }

    public ICollection<IGraphElement> TabulatingElements {
      get { return tabulatingElements; }
    }

    public ICollection<IGraphElement> SelectedElements {
      get {
        var selected = new List<IGraphElement>(tabulatingListView.SelectedItems.Count);

        foreach(ListViewItem item in tabulatingListView.SelectedItems) {
          if (item is ElementListViewItem)
            selected.Add((item as ElementListViewItem).Element);
        }

        return selected;
      }
    }

    public bool CompareEntryType {
      get { return compareEntryType; }
      set {
        if (value == compareEntryType)
          return;

        compareEntryType = value;

        Sort(entryCompareMode, entryCompareOrder);
      }
    }

    private bool IsSelectedItemChanged {
      get {
        foreach (ListViewItem item in tabulatingListView.SelectedItems) {
          if (!lastSelectedItems.Contains(item))
            return true;
        }

        return false;
      }
    }

    public Tabulator(ListView tabulatingListView, TextBox tabulatingTitleTextBox)
    {
      this.tabulatingListView = tabulatingListView;
      this.tabulatingTitleTextBox = tabulatingTitleTextBox;

      InitializeListView();

      InitializeHandler();
    }

    private void InitializeListView()
    {
      tabulatingListView.Activation = ItemActivation.Standard;
      tabulatingListView.AllowColumnReorder = true;
      tabulatingListView.CheckBoxes = false;
      tabulatingListView.HeaderStyle = ColumnHeaderStyle.Clickable;
      tabulatingListView.View = View.Details;
      tabulatingListView.FullRowSelect = true;
      tabulatingListView.MultiSelect = true;
      tabulatingListView.Scrollable = true;
      tabulatingListView.LabelEdit = false;
      tabulatingListView.ListViewItemSorter = null;
      tabulatingListView.ShowItemToolTips = true;

      tabulatingListView.Columns.Clear();
      tabulatingListView.Columns.Add(new EntryListViewColumn("名前", 120, DirectoryEntryCompareMode.ByName));
      tabulatingListView.Columns.Add(new EntryListViewColumn("種類", 32, DirectoryEntryCompareMode.ByExtension));
      tabulatingListView.Columns.Add(new EntryListViewColumn("容量[%]", 64, DirectoryEntryCompareMode.BySize));
      tabulatingListView.Columns.Add(new EntryListViewColumn("容量[Bytes]", 64, DirectoryEntryCompareMode.BySize));
      tabulatingListView.Columns.Add(new EntryListViewColumn("フルパス", 540, DirectoryEntryCompareMode.ByFullName));

      tabulatingTitleTextBox.ReadOnly = true;
    }

    private void InitializeHandler()
    {
      tabulatingListView.MouseUp += delegate(object sender, MouseEventArgs e) {
        var item = tabulatingListView.GetItemAt(e.X, e.Y) as ElementListViewItem;

        if (item != null)
          OnItemClicked(new TabulatorMouseEventArgs(item.Element, e));

        // 選択されているアイテムが変わった場合
        if (IsSelectedItemChanged)
          // SelectedItemChangedイベントは、アイテムの内容が変わる度に発生するので使わない
          // (複数選択→1件選択だと、減った選択数分SelectedItemChangedイベントが発生する)
          // キー入力での選択に対応できないことについては、割り切る
          OnSelectedElementChanged(EventArgs.Empty);
      };

      tabulatingListView.ColumnClick += delegate(object sender, ColumnClickEventArgs e) {
        var column = (sender as ListView).Columns[e.Column] as EntryListViewColumn;

        if (column == null)
          return;

        if (column.CompareMode == entryCompareMode) {
          // ソート方法が同じ場合は、昇降順を逆にする
          if (entryCompareOrder == DirectoryEntryCompareOrder.Ascending)
            Sort(entryCompareMode, DirectoryEntryCompareOrder.Descending);
          else
            Sort(entryCompareMode, DirectoryEntryCompareOrder.Ascending);
        }
        else {
          // ソート方法が異なる場合は、新しいソート方法で並べ替える
          Sort(column.CompareMode, DirectoryEntryCompareOrder.Ascending);
        }
      };
    }

    public void SetElements(ICollection<IGraphElement> elements, string title)
    {
      tabulatingElements = elements;

      Sort(entryCompareMode, entryCompareOrder);

      tabulatingTitleTextBox.Text = title;
    }

    private void Sort(DirectoryEntryCompareMode sortMode, DirectoryEntryCompareOrder sortOrder)
    {
      if (tabulatingElements == null)
        return;

      // 指定の方法で並び替える
      var sortedTabulatingElements = new List<IGraphElement>(tabulatingElements);
      var comparison = DirectoryEntryComparison.GetComparison(sortMode, sortOrder, compareEntryType);

      if (comparison != null) {
        sortedTabulatingElements.Sort(delegate(IGraphElement x, IGraphElement y) {
          return comparison(x.Entry, y.Entry);
        });
      }

      tabulatingElements = sortedTabulatingElements;

      UpdateListItem();

      entryCompareMode = sortMode;
      entryCompareOrder = sortOrder;
    }

    private void UpdateListItem()
    {
      Console.WriteLine("Updating tabulation items...");

      var stopwatch = Stopwatch.StartNew();

      // tabulateMaxで指定されている件数に絞る
      var viewingElements = (0 == tabulateMax)
        ? tabulatingElements
        : tabulatingElements.Take(tabulateMax);

      // ListViewの項目を作成する
      try {
        tabulatingListView.BeginUpdate();

        tabulatingListView.Items.Clear();

        foreach (var element in viewingElements) {
          tabulatingListView.Items.Add(ElementListViewItem.Create(element));
        }

        if (tabulatingElements.Count == 0) {
          tabulatingListView.Items.Add(new ListViewItem("該当するファイル・フォルダはありません"));
        }
        else {
          var remain = tabulatingElements.Count - viewingElements.Count();

          if (0 < remain)
            tabulatingListView.Items.Add(new ListViewItem(string.Format("他に{0}個のファイル・フォルダがあります", remain)));
        }
      }
      finally {
        tabulatingListView.EndUpdate();
      }

      stopwatch.Stop();

      Console.WriteLine("  Elaplsed: {0}", stopwatch.Elapsed);
      Console.WriteLine();
    }

    protected virtual void OnItemClicked(TabulatorMouseEventArgs e)
    {
      var ev = ItemClicked;

      if (ev != null)
        ev(this, e);
    }

    protected virtual void OnSelectedElementChanged(EventArgs e)
    {
      var ev = SelectedElementChanged;

      if (ev != null)
        ev(this, e);
    }

    private ListView tabulatingListView = null;
    private TextBox tabulatingTitleTextBox = null;
    // TODO: IReadOnlyColleciton
    private ICollection<IGraphElement> tabulatingElements = null;
    private int tabulateMax = 30;
    private DirectoryEntryCompareMode entryCompareMode = DirectoryEntryCompareMode.BySize;
    private DirectoryEntryCompareOrder entryCompareOrder = DirectoryEntryCompareOrder.Ascending;
    private bool compareEntryType = true;
    private List<ListViewItem> lastSelectedItems = new List<ListViewItem>(1);
  }
}
