//  
//  Author:
//       smdn <smdn@smdn.jp>
// 
//  Copyright (c) 2003-2013 smdn
// 
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.Drawing.Text;
using System.Text.RegularExpressions;
using System.Windows.Forms;

using Smdn.IO.FileSystems;

namespace Smdn.Applications.BittenBaum.UI {
  using Analysis;
  using Graphing;
  using Rendering;

  public class Viewer {
    /*
     * inner types
     */
    private class ViewingHistoryEntry {
      public Graph Graph {
        get; set;
      }

      public DirectoryTree DirectoryTree {
        get { return Graph.DirectoryTree; }
      }

      public Bitmap GraphImage {
        get; set;
      }

      public ViewingHistoryEntry(Graph graph, Bitmap graphImage)
      {
        this.Graph = graph;
        this.GraphImage = graphImage;
      }
    }

    private class ViewingHistory {
      public ViewingHistoryEntry Current {
        get
        {
          if (currentEntry < 0) {
            return null;
          }
          else {
            if (entries.Count == 0)
              return null;
            else
              return entries[currentEntry];
          }
        }
      }

      public ViewingHistory(int historyMax)
      {
        this.historyMax = historyMax;
        this.entries = new List<ViewingHistoryEntry>(historyMax);

        Clear();
      }

      public void Add(ViewingHistoryEntry entry)
      {
        if (entries.Count == historyMax) {
          // ヒストリの最大数を超える場合は、一番古いヒストリを削除する
          var oldestEntry = entries[entries.Count - 1];

          entries.Remove(oldestEntry);

          oldestEntry.GraphImage.Dispose();
        }

        // Backしている状態でAddするときは、Forward分を削除する
        // (以後Forwardできないようにする)
        for (; 0 < currentEntry; currentEntry--) {
          entries.RemoveAt(0);
        }

        currentEntry = 0;

        entries.Insert(currentEntry, entry);
      }

      public void Clear()
      {
        Clear(false);
      }

      public void Clear(bool exceptCurrent)
      {
        ViewingHistoryEntry current = null;

        if (exceptCurrent) {
          current = this.Current;

          entries.Remove(current);
        }

        foreach (var entry in entries) {
          entry.GraphImage.Dispose();
        }

        entries.Clear();

        currentEntry = -1;

        if (current != null)
          Add(current);
      }

      public void Forward()
      {
        if (0 == entries.Count)
          currentEntry = -1;

        if (0 < currentEntry)
          currentEntry -= 1;
      }

      public void Back()
      {
        if (0 == entries.Count)
          currentEntry = -1;

        if (currentEntry < entries.Count - 1)
          currentEntry += 1;
      }

      public void GetHistory(out ViewingHistoryEntry[] forwards, out ViewingHistoryEntry[] backwards)
      {
        if (0 == entries.Count) {
          forwards = EmptyArray<ViewingHistoryEntry>.Instance;
          backwards = EmptyArray<ViewingHistoryEntry>.Instance;
          return;
        }

        forwards  = new ViewingHistoryEntry[currentEntry];
        backwards = new ViewingHistoryEntry[entries.Count - (currentEntry + 1)];

        entries.CopyTo(0, forwards, 0, forwards.Length);
        entries.CopyTo(currentEntry, backwards, 0, backwards.Length);

        Array.Reverse(forwards);
      }

      public void GetHistory(out DirectoryTree[] forwards, out DirectoryTree[] backwards)
      {
        ViewingHistoryEntry[] forwardEntries, backwardEntries;

        GetHistory(out forwardEntries, out backwardEntries);

        forwards = Array.ConvertAll(forwardEntries, delegate(ViewingHistoryEntry entry) {
          return entry.DirectoryTree;
        });

        backwards = Array.ConvertAll(backwardEntries, delegate(ViewingHistoryEntry entry) {
          return entry.DirectoryTree;
        });
      }

      private int historyMax = 0;
      private int currentEntry = -1;
      private List<ViewingHistoryEntry> entries = null;
    }

    /*
     * instance members
     */
    public event EventHandler<ViewerMouseEventArgs> ElementAtCursorChanged = null;
    public event EventHandler<ViewerMouseEventArgs> ElementClicked = null;
    public event MouseEventHandler MouseClicked = null;
    public event EventHandler<ViewerStatusEventArgs> StatusChanged = null;
    public event EventHandler GraphRootChanged = null;
    public event EventHandler HistoryChanged = null;

    private const int viewingHistoryMax = 5;

    private Visualizer visualizer = null;
    private Control viewingControl = null;
    private IGraphElement elementAtCursor = null;
    private Renderer elementRetrievingRenderer = null;
    private PartialGraphImage highlightedPartialGraph = null;
    private float viewingGraphRatio = 0.0f;
    private RectangleF viewingGraphBounds = Rectangle.Empty;
    private ViewingHistory viewingHistory = new ViewingHistory(viewingHistoryMax);
    private InterpolationMode interpolationMode = InterpolationMode.Default;

    public Visualizer Visualizer {
      get {
        return visualizer;
      }
    }

    public Control ViewingControl {
      get {
        return viewingControl;
      }
    }

    public Graph Graph {
      get {
        if (viewingHistory.Current == null)
          return null;
        else
          return viewingHistory.Current.Graph;
      }
    }

    /// <summary>
    /// コンストラクタ
    /// </summary>
    public Viewer(Control viewingControl)
    {
      this.viewingControl = viewingControl;

      InitializeHandler();
    }

    private void InitializeHandler()
    {
      // リサイズ時に描画する
      viewingControl.Resize += delegate(object sender, EventArgs e) {
        MeasureViewingSize();

        viewingControl.Invalidate();
      };

      // 描画処理
      viewingControl.Paint += delegate(object sender, PaintEventArgs e) {
        RenderView(e.Graphics);
      };

      // マウスの移動
      viewingControl.MouseMove += delegate(object sender, MouseEventArgs e) {
        var point = new Point(e.X, e.Y);
        var element = GetElementAt(point);

        if (elementAtCursor != element) {
          elementAtCursor = element;

          if (element == null)
            OnEntryAtCursorChanged(new ViewerMouseEventArgs(null, e));
          else
            OnEntryAtCursorChanged(new ViewerMouseEventArgs(element, e));
        }
      };

      // 左右クリック
      viewingControl.MouseClick += delegate(object sender, MouseEventArgs e) {
        if (elementAtCursor == null)
          OnMouseClicked(new MouseEventArgs(e.Button, e.Clicks, e.X, e.Y, e.Delta));
        else
          OnEntryClicked(new ViewerMouseEventArgs(elementAtCursor, e));
      };
    }

    /// <summary>
    /// ツリーを見る
    /// </summary>
    /// <param name="tree">見る対象のツリー</param>
    /// <param name="updateGraph">ツリーを表示する前にグラフを再作成するかどうか</param>
    /// <param name="updateRendered">ツリーを表示する前にグラフを再描画するかどうか</param>
    /// <param name="addToHistory">作成・描画したグラフをヒストリに追加するかどうか</param>
    /// <param name="updateCurrentHistory">ヒストリに追加しない場合、現在のヒストリの状態を更新するかどうか</param>
    private void ViewTree(DirectoryTree tree, bool updateGraph, bool updateRendered, bool addToHistory, bool updateCurrentHistory)
    {
      var stopwatch = Stopwatch.StartNew();

      Console.WriteLine("Selected tree '{0}'", tree.Directory.FullName);

      // ハイライトされた部分は無効となるため破棄する
      if (highlightedPartialGraph != null) {
        lock (highlightedPartialGraph) {
          highlightedPartialGraph.Dispose();
          highlightedPartialGraph = null;
        }
      }

      var bakingBaumkuchen = visualizer.RenderingContext.Parameter.ColoringMode == ColoringMode.DirectoryBaumkuchen;

      // グラフの再作成・再描画
      interpolationMode = visualizer.RenderingContext.Parameter.InterpolationMode;

      if (updateGraph || updateRendered) {
        OnStatusChanged(new ViewerStatusEventArgs(bakingBaumkuchen ? "バウムクーヘンを焼いています..." : "グラフを描画しています..."));

        // グラフを再作成する
        if (updateGraph) {
          visualizer.GraphingContext.Parameter.DirectoryTree = tree;

          visualizer.Graph(false);

          // 位置取得用のRendererは無効となるため再作成させる
          elementRetrievingRenderer = null;
        }

        // グラフを再描画する
        if (updateRendered) {
          visualizer.RenderingContext.Parameter.Graph = visualizer.GraphingContext.Result.Graph;

          visualizer.Render(false);
        }
      }

      // 位置取得用のRendererを再作成
      if (elementRetrievingRenderer == null)
        elementRetrievingRenderer = new Renderer(visualizer.RenderingContext);

      if (addToHistory) {
        // 新しいエントリをヒストリに追加
        viewingHistory.Add(new ViewingHistoryEntry(
          visualizer.GraphingContext.Result.Graph,
          visualizer.RenderingContext.Result.GraphImage));

        OnHistoryChanged(EventArgs.Empty);
      }
      else if (updateCurrentHistory) {
        // 現在のヒストリの状態を更新
        lock(viewingHistory.Current) {
          viewingHistory.Current.Graph = visualizer.GraphingContext.Result.Graph;
          viewingHistory.Current.GraphImage.Dispose();
          viewingHistory.Current.GraphImage = visualizer.RenderingContext.Result.GraphImage;
        }
      }

      // グラフのルートの変更を通知
      OnGraphRootChanged(EventArgs.Empty);

      // 再描画
      MeasureViewingSize();

      viewingControl.Invalidate();

      // ステータスを更新
      stopwatch.Stop();

      OnStatusChanged(new ViewerStatusEventArgs(
        string.Format("{0} ({1:N2}秒, {2})",
          bakingBaumkuchen ? "バウムクーヘンが焼けました" : "描画が完了しました",
          stopwatch.Elapsed.TotalSeconds,
          tree.Directory.FullName)
        ));
    }

    /// <summary>
    /// ヒストリに設定されているツリーを見る
    /// </summary>
    /// <param name="entry">ヒストリのエントリ</param>
    /// <param name="updateGraph">グラフを再作成するかどうか</param>
    /// <param name="updateRendered">グラフを再描画するかどうか</param>
    private void ViewHistory(ViewingHistoryEntry entry, bool updateGraph, bool updateRendered)
    {
      // 各タスクのパラメータを復元
      visualizer.GraphingContext.Parameter.DirectoryTree = viewingHistory.Current.DirectoryTree;
      visualizer.RenderingContext.Parameter.Graph = viewingHistory.Current.Graph;

      if (updateGraph || updateRendered) {
        ViewTree(entry.DirectoryTree, updateGraph, updateRendered, false, true);
      }
      else {
        // 状態を更新しない場合は各タスクの結果も復元
        visualizer.GraphingContext.Result.Graph = viewingHistory.Current.Graph;
        visualizer.RenderingContext.Result.GraphImage = viewingHistory.Current.GraphImage;

        ViewTree(entry.DirectoryTree, false, false, false, false);
      }

      OnHistoryChanged(EventArgs.Empty);
    }

    /// <summary>
    /// グラフの描画領域を再計算する
    /// </summary>
    private void MeasureViewingSize()
    {
      if (viewingHistory.Current == null)
         return;

      // グラフ全体を描画できる拡大縮小率と、描画領域を求める
      var graphImage = viewingHistory.Current.GraphImage;
      var tanGraphImage = graphImage.Height / (float)graphImage.Width;
      var tanPictureBox = viewingControl.Height / (float)viewingControl.Width;

      if (tanGraphImage < tanPictureBox) {
        viewingGraphRatio = viewingControl.Width / (float)graphImage.Width;

        viewingGraphBounds = new RectangleF(0.0f, 0.0f, viewingControl.Width, graphImage.Height * viewingGraphRatio);
        viewingGraphBounds.Y = (viewingControl.Height - graphImage.Height * viewingGraphRatio) * 0.5f;
      }
      else {
        viewingGraphRatio = viewingControl.Height / (float)graphImage.Height;

        viewingGraphBounds = new RectangleF(0.0f, 0.0f, graphImage.Width * viewingGraphRatio, viewingControl.Height);
        viewingGraphBounds.X = (viewingControl.Width - graphImage.Width * viewingGraphRatio) * 0.5f;
      }
    }

    /// <summary>
    /// グラフの描画を行う
    /// </summary>
    private void RenderView(Graphics g)
    {
      g.Clear(SystemColors.Control);

      if (viewingHistory.Current == null)
         return;

      g.InterpolationMode = interpolationMode;

      // 描画する
      switch (viewingHistory.Current.Graph.Type) {
        case GraphType.Disc:
          g.DrawImage(viewingHistory.Current.GraphImage, viewingGraphBounds);
          break;

        default:
          break;
      }

      // ハイライトされた部分を上書きする
      if (highlightedPartialGraph != null) {
        g.DrawImage(
          highlightedPartialGraph.Image,
          viewingGraphBounds.X + highlightedPartialGraph.ImageOffsetX * viewingGraphRatio,
          viewingGraphBounds.Y + highlightedPartialGraph.ImageOffsetY * viewingGraphRatio,
          highlightedPartialGraph.Image.Width * viewingGraphRatio,
          highlightedPartialGraph.Image.Height * viewingGraphRatio);
      }
    }

    /// <summary>
    /// 新しいVisualizerを設定する
    /// </summary>
    public void SetVisualizer(Visualizer visualizer)
    {
      if (this.visualizer == visualizer)
         return;

      this.visualizer = visualizer;

      // 位置取得用のRendererは無効となるため再作成させる
      elementRetrievingRenderer = null;

      // ヒストリをクリアする
      viewingHistory.Clear();

      ViewTree(visualizer.GraphingContext.Parameter.DirectoryTree, false, false, true, false);
    }

    /// <summary>
    /// ヒストリを進む
    /// </summary>
    public void ForwardHistory(int step)
    {
      if (step <= 0)
        throw ExceptionUtils.CreateArgumentMustBeNonZeroPositive("step", step);
      if (visualizer == null)
        return;

      for (var s = 0; s < step; s++) {
        viewingHistory.Forward();
      }

      if (viewingHistory.Current != null)
        ViewHistory(viewingHistory.Current, false, false);
    }

    /// <summary>
    /// ヒストリを遡る
    /// </summary>
    public void BackHistory(int step)
    {
      if (step <= 0)
        throw ExceptionUtils.CreateArgumentMustBeNonZeroPositive("step", step);
      if (visualizer == null)
        return;

      for (var s = 0; s < step; s++) {
        viewingHistory.Back();
      }

      if (viewingHistory.Current != null)
        ViewHistory(viewingHistory.Current, false, false);
    }

    /// <summary>
    /// ヒストリを取得する
    /// </summary>
    public void GetHistory(out DirectoryTree[] forwards, out DirectoryTree[] backwards)
    {
      if (visualizer == null) {
        forwards = EmptyArray<DirectoryTree>.Instance;
        backwards = EmptyArray<DirectoryTree>.Instance;
        return;
      }

      viewingHistory.GetHistory(out forwards, out backwards);
    }

    /// <summary>
    /// 指定された要素をハイライトする
    /// </summary>
    public void HighlightElements(ICollection<IGraphElement> elements)
    {
      if (elements == null)
        return;

      // ハイライトされた部分は無効となるため破棄する
      if (null != highlightedPartialGraph) {
        lock (highlightedPartialGraph) {
          highlightedPartialGraph.Dispose();
          highlightedPartialGraph = null;
        }
      }

      if (0 < elements.Count) {
        Console.WriteLine("Highlighting {0} elements...", elements.Count);

        var stopwatch = Stopwatch.StartNew();

        // ハイライトしたイメージを取得する
        highlightedPartialGraph = (new Renderer(visualizer.RenderingContext)).GetElementImage(elements, true);

        stopwatch.Stop();

        Console.WriteLine("  Elaplsed: {0}", stopwatch.Elapsed);
        Console.WriteLine();
      }

      // 再描画
      viewingControl.Invalidate();
    }

    /// <summary>
    /// ハイライトをクリアする
    /// </summary>
    public void ClearHighlight()
    {
      if (null != highlightedPartialGraph) {
        lock (highlightedPartialGraph) {
          highlightedPartialGraph.Dispose();
          highlightedPartialGraph = null;
        }
      }

      // 再描画
      viewingControl.Invalidate();
    }

    /// <summary>
    /// 指定されたツリーを見る
    /// </summary>
    public void ViewTree(DirectoryTree tree)
    {
      if (viewingHistory.Current != null && viewingHistory.Current.DirectoryTree == tree)
        return;

      ViewTree(tree, true, true, true, false);
    }

    /// <summary>
    /// グラフ上の任意の点に描画されているエントリを取得する
    /// </summary>
    private IGraphElement GetElementAt(Point point)
    {
      if (visualizer == null)
         return null;

      var parameter = visualizer.RenderingContext.Parameter;
      var elementPoint = PointF.Empty;

      switch (viewingHistory.Current.Graph.Type) {
        case GraphType.Disc:
          // 現在の描画拡大縮小率を求める
          var ratio = (float)Math.Max(parameter.OutputWidth / (float)viewingControl.Width,
                                       parameter.OutputHeight / (float)viewingControl.Height);

          // 拡大縮小率を元に、元のサイズでの座標に変換する
          var imageCenter = new PointF(viewingControl.Width * 0.5f, viewingControl.Height * 0.5f);
          var discCenter = new PointF(parameter.OutputWidth * 0.5f, parameter.OutputHeight * 0.5f);

          elementPoint = new PointF(discCenter.X + (point.X - imageCenter.X) * ratio,
                                     discCenter.Y + (point.Y - imageCenter.Y) * ratio);

          break;

        default:
          return null;
      }

      // 特定の位置にあるエントリを取得する
      return elementRetrievingRenderer.GetElementAt(elementPoint);
    }

    /// <summary>
    /// 現在の状態を破棄して更新する
    /// </summary>
    public void DiscardView(bool updateGraph)
    {
      if (viewingHistory.Current == null)
        return;

      ViewHistory(viewingHistory.Current,
                  updateGraph,
                  true);

      // 描画済みの画像を参照させないよう、現在のヒストリ以外をクリアする
      viewingHistory.Clear(true);

      OnHistoryChanged(EventArgs.Empty);
    }

    protected virtual void OnEntryAtCursorChanged(ViewerMouseEventArgs e)
    {
      var ev = ElementAtCursorChanged;

      if (ev != null)
        ev(this, e);
    }

    protected virtual void OnEntryClicked(ViewerMouseEventArgs e)
    {
      var ev = ElementClicked;

      if (ev != null)
        ev(this, e);
    }

    protected virtual void OnMouseClicked(MouseEventArgs e)
    {
      var ev = MouseClicked;

      if (ev != null)
        ev(this, e);
    }

    protected virtual void OnStatusChanged(ViewerStatusEventArgs e)
    {
      var ev = StatusChanged;

      if (ev != null)
        ev(this, e);
    }

    protected virtual void OnGraphRootChanged(EventArgs e)
    {
      var ev = GraphRootChanged;

      if (ev != null)
        ev(this, e);
    }

    protected virtual void OnHistoryChanged(EventArgs e)
    {
      var ev = HistoryChanged;

      if (ev != null)
        ev(this, e);
    }
  }
}
