//  
//  Author:
//       smdn <smdn@smdn.jp>
// 
//  Copyright (c) 2003-2013 smdn
// 
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Reflection;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;
#if GUI
using System.Windows.Forms;
#endif

using Smdn.IO.FileSystems;

namespace Smdn.Applications.BittenBaum {
  using Analysis;
  using Formatting;
  using Graphing;
  using Rendering;

  public partial class BittenBaum {
    private enum RunMode {
      Default,
      Visualize,
      Format,
      Usage,
      Abort,
    }

    public static string ApplicationName {
      get {
        return (Attribute.GetCustomAttribute(Assembly.GetExecutingAssembly(),
                                             typeof(AssemblyProductAttribute)) as AssemblyProductAttribute).Product;
      }
    }

    public static string ApplicationVersion {
      get {
        return (Attribute.GetCustomAttribute(Assembly.GetExecutingAssembly(),
                                             typeof(AssemblyFileVersionAttribute)) as AssemblyFileVersionAttribute).Version;
      }
    }

    public static string VersionedApplicationName {
      get {
        var product = (Attribute.GetCustomAttribute(Assembly.GetExecutingAssembly(),
                                                    typeof(AssemblyProductAttribute)) as AssemblyProductAttribute).Product;

        return string.Concat(product, " version ", ApplicationVersion);
      }
    }

    public static string ApplicationCopyright {
      get {
        return (Attribute.GetCustomAttribute(Assembly.GetExecutingAssembly(),
                                             typeof(AssemblyCopyrightAttribute)) as AssemblyCopyrightAttribute).Copyright;
      }
    }

    public static string ApplicationDistributionUrl {
      get {
        return (Attribute.GetCustomAttribute(Assembly.GetExecutingAssembly(),
                                             typeof(AssemblyDistributionPageAttribute)) as AssemblyDistributionPageAttribute).Url;
      }
    }

    [STAThread]
    public static void Main(string[] args)
    {
      var runmode = RunMode.Default;
      Visualizer visualizer = null;
      string outputFile = null;

#if GUI
      if (!Runtime.IsRunningOnWindows)
        runmode = ParseArguments(args, out visualizer, out outputFile);
#else
      runmode = ParseArguments(args, out visualizer, out outputFile);
#endif

      switch(runmode) {
        case RunMode.Usage:
          // 使い方
          ShowUsage();
          return;

        case RunMode.Abort:
          // 中断
          Console.Error.WriteLine("中断しました。");
          return;

        case RunMode.Default:
#if GUI
          if (Runtime.IsRunningOnWindows) {
#if !DEBUG
            // Windows上ではログをファイルに書き出す
            TextWriter writer = null;

            try {
              writer = new StreamWriter(System.Reflection.Assembly.GetExecutingAssembly().GetName().Name + ".log");

              (writer as StreamWriter).AutoFlush = true;

              writer = TextWriter.Synchronized(writer);

              Console.SetOut(writer);
            }
            catch {
              // ignore exceptions
            }
#endif

            Console.SetError(Console.Out);
          }

          // デフォルト動作
          try {
            Config.Load(Path.GetDirectoryName(Application.ExecutablePath));

            Application.EnableVisualStyles();
            Application.SetCompatibleTextRenderingDefault(false);
            Application.Run(new UI.MainWindow());
          }
          finally {
            Config.Save();
          }
          break;
#else
          ShowUsage();
          return;
#endif

        case RunMode.Format:
        case RunMode.Visualize:
          // コマンドラインでの解析
          Console.WriteLine("Analysing directory: {0}", visualizer.AnalysisContext.Parameter.AnalysisDirectory.FullName);

          // ディレクトリを解析し、グラフを作成する
          var t = new Thread((ThreadStart)delegate {
            if (runmode == RunMode.Visualize)
              visualizer.Visualize();
            else
              visualizer.Format();
          });

          t.Start();

          for (; !t.Join(125); ) {
            // 進捗を表示
            var status = visualizer.Status;

            if (status != null) {
              if (status.Total == -1)
                ConsoleUtils.WriteProgress(status.Percentage);
              else
                ConsoleUtils.WriteProgress((int)status.Processed, (int)status.Total);
            }
          }

          // 描画/整形したグラフを保存する
          if (runmode == RunMode.Visualize)
            visualizer.RenderingContext.Result.Save(outputFile);

          Console.WriteLine("The graph was output to: {0}", outputFile);
          Console.WriteLine("Completed.");

          break;
      } // switch(runmode)
    }

    private static RunMode ParseArguments(string[] args, out Visualizer visualizer, out string outputFile)
    {
      visualizer = null;
      outputFile = null;

      // 必須オプションがない場合は、デフォルト動作
      if (args.Length < 1)
        return RunMode.Default;

      foreach (var arg in args) {
        // 使い方を表示
        switch (arg) {
          case "/?":
            case "/help":
            return RunMode.Usage;

            case "--help":
            if (Runtime.IsRunningOnWindows)
              break; // do nothing
            else
              return RunMode.Usage;
        }
      }

      var analysisDirectory = args[0];

      if (Runtime.IsRunningOnWindows &&
          Regex.IsMatch(analysisDirectory, @"^[A-Z]\:$", RegexOptions.CultureInvariant | RegexOptions.IgnoreCase))
        // ドライブレターのみの場合はディレクトリ区切り文字を追加する
        analysisDirectory += Path.DirectorySeparatorChar;

      // 指定されたディレクトリが無い場合は使い方を表示して終了
      if (!System.IO.Directory.Exists(analysisDirectory)) {
        Console.Error.WriteLine("ディレクトリ{0}は存在しません。", analysisDirectory);

        return RunMode.Abort;
      }

      // 解析されていない引数
      var argsRemain = new List<string>(args);

      argsRemain.RemoveAt(0);

      // 引数を解析
      RunMode runMode = RunMode.Visualize;
      int? outputWidth = null;
      int? outputHeight = null;
      int? graphingDepth = null;
      float? renderingInnerDiameter = null;
      ColoringMode? coloringMode = null;
      int? coloringHue = null;
      Regex excludeRegex = null;
      FormatType? formatType = null;
      bool? silent = null;
      bool? compareEntryType = null;
      DirectoryEntryCompareMode? entryCompareMode = null;
      DirectoryEntryCompareOrder? entryCompareOrder = null;
      bool? calculateRangeByDriveUsage = null;
      Color? backgroundColor = null;
      Color? textColor = null;
      int? fontSize = null;
      bool? drawSmallEntries = null;
      bool? drawNames = null;
      bool? enableGradation = null;

      foreach (var a in args) {
        if (!a.StartsWith('/'))
           continue;

        // '/'削除、小文字化、値分割
        var arg = a.Substring(1).ToLowerInvariant();
        var val = string.Empty;

        if (arg.Contains(":")) {
          val = a.Substring(arg.IndexOf(":") + 2);
          arg = arg.Substring(0, arg.IndexOf(":"));
        }

        switch (arg) {
          case "width":
          case "w":
            // 出力ファイルの幅
            outputWidth = int.Parse(val);
            break;

          case "height":
          case "h":
            // 出力ファイルの高さ
            outputHeight = int.Parse(val);
            break;

          case "depth":
          case "d":
            // 最大の深さ
            graphingDepth = int.Parse(val);
            break;

          case "dia":
            // グラフの内径
            renderingInnerDiameter = float.Parse(val);

            if (renderingInnerDiameter < 0.1f || 0.9f < renderingInnerDiameter)
              renderingInnerDiameter = null;

            break;

          case "colmode":
            // グラフの色付けモード
            coloringMode = EnumUtils.ParseIgnoreCase<ColoringMode>(val);
            break;

          case "colhue":
            // グラフの色相
            coloringHue = int.Parse(val);
            break;

          case "exclude":
            // 除外ディレクトリの正規表現
            excludeRegex = new Regex(val);
            break;

          case "xml":
            // XML形式で出力
            formatType = FormatType.Xml;
            break;

          case "csv":
            // CSV形式で出力
            formatType = FormatType.Csv;
            break;

          case "silent":
          case "quiet":
          case "s":
            // サイレント
            silent = true;
            break;

          case "individuate":
          case "i":
            // ファイルとディレクトリを区別
            compareEntryType = true;
            break;

          case "name":
          case "n":
            // ファイル名で並べ替え
            entryCompareMode = DirectoryEntryCompareMode.ByName;
            break;

          case "fullpath":
          case "f":
            // フルパス名で並べ替え
            entryCompareMode = DirectoryEntryCompareMode.ByFullName;
            break;

          case "reverse":
            // 逆順え並べ替え
            entryCompareOrder = DirectoryEntryCompareOrder.Descending;
            break;

          case "bg":
            // 背景色
            backgroundColor = Color.FromKnownColor((KnownColor)Enum.Parse(typeof(KnownColor), val, true));
            break;

          case "text":
            // 文字色
            textColor = Color.FromKnownColor((KnownColor)Enum.Parse(typeof(KnownColor), val, true));
            break;

          case "textsize":
            // 文字サイズ
            fontSize = int.Parse(val);
            break;

          case "driveusage":
            // ドライブ総容量に対する占有率で描画する
            calculateRangeByDriveUsage = true;
            break;

          case "detail":
            // 微少なエントリを描画する
            drawSmallEntries = true;
            break;

          case "nogradation":
            // グラデーションを無効にする
            enableGradation = false;
            break;

          case "nolabel":
            // エントリ名を描画しない
            drawNames = false;
            break;

          default:
            continue;
        }

        // 解析済み
        argsRemain.Remove(a);
      }

      // 出力ファイル名
      if (argsRemain.Count == 0) {
        // 省略された場合は、カレントディレクトリに書き出す
        if (formatType.HasValue) {
          switch (formatType.Value) {
            case FormatType.Xml:
              outputFile = "graph.xml";
              break;

              case FormatType.Csv:
              outputFile = "graph.csv";
              break;
          }
        }
        else {
          outputFile = "graph.png";
        }
      }
      else {
        // 解析されなかった引数のうち、一番最後をファイル名と見なす
        outputFile = argsRemain[argsRemain.Count - 1];
      }

      // 出力先ディレクトリの存在を確認し、必要なら作成する
      outputFile = Path.GetFullPath(outputFile);

      var outputDirectory = Path.GetDirectoryName(outputFile);

      if (!System.IO.Directory.Exists(outputDirectory)) {
        if (ConsoleUtils.AskYesNo(false, "出力先のディレクトリ'{0}'が存在しません。　作成しますか", outputDirectory))
          System.IO.Directory.CreateDirectory(outputDirectory);
        else
          return RunMode.Abort;
      }

      // Visualizerのインスタンスを作成
      visualizer = new Visualizer(FileSystem.GetDirectory(analysisDirectory));

      /*
       * AnalysisContextのパラメータを設定
       */
      var analysisParameter = visualizer.AnalysisContext.Parameter;

      if (excludeRegex == null)
        analysisParameter.ScanPredicate = null;
      else
        analysisParameter.ScanPredicate = delegate(DirectoryEntry entry) {
          return !(entry is Smdn.IO.FileSystems.Directory && excludeRegex.IsMatch(entry.FullName));
        };

      /*
       * GraphingContextのパラメータを設定
       */
      var graphingParameter = visualizer.GraphingContext.Parameter;

      if (graphingDepth.HasValue)
        graphingParameter.MaxDepth = graphingDepth.Value;

      if (calculateRangeByDriveUsage.HasValue)
        graphingParameter.CalculateRangeByDriveUsage = calculateRangeByDriveUsage.Value;

      graphingParameter.Comparison = DirectoryEntryComparison.GetComparison(entryCompareMode ?? DirectoryEntryCompareMode.BySize,
                                                                            entryCompareOrder ?? DirectoryEntryCompareOrder.Ascending,
                                                                            compareEntryType ?? false);

      //graphingParameter.Type = GraphType.Disc;

      /*
       * RenderingContextのパラメータを設定
       */
      var renderingParameter = visualizer.RenderingContext.Parameter;

      if (outputWidth.HasValue)
        renderingParameter.OutputWidth = outputWidth.Value;

      if (outputHeight.HasValue)
        renderingParameter.OutputHeight = outputHeight.Value;

      if (renderingInnerDiameter.HasValue)
        renderingParameter.InnerDiameterPercent = renderingInnerDiameter.Value;

      if (coloringMode.HasValue)
        renderingParameter.ColoringMode = coloringMode.Value;

      if (coloringHue.HasValue)
        renderingParameter.ColoringHue = coloringHue.Value;

      renderingParameter.DrawSmallEntries = drawSmallEntries ?? false;

      if (drawNames.HasValue)
        renderingParameter.DrawNames = drawNames.Value;

      renderingParameter.EnableGradation = enableGradation ?? true;

      if (backgroundColor.HasValue)
        renderingParameter.BackgroundColor = backgroundColor.Value;

      if (textColor.HasValue)
        renderingParameter.TextColor = textColor.Value;

      if (fontSize.HasValue)
        renderingParameter.FontSize = fontSize.Value;

      /*
       * FormattingContextのパラメータを設定
       */
      var formattingParameter = visualizer.FormattingContext.Parameter;

      if (formatType.HasValue) {
        formattingParameter.Type = formatType.Value;

        // 出力フォーマットが指定されている場合は、RunModeをFormatにする
        runMode = RunMode.Format;

        // 出力用のStreamを設定する
        formattingParameter.OutputStream = new FileStream(outputFile, FileMode.Create, FileAccess.Write, FileShare.None);
        formattingParameter.CloseStreamAfterFormatting = true;
      }

      // silentの場合は標準出力をTextWriter.Nullにリダイレクトする
      if (silent ?? false)
        Console.SetOut(System.IO.TextWriter.Null);

      return runMode;
    }

    /// <summary>
    /// 使い方の表示
    /// </summary>
    private static void ShowUsage()
    {
      var renderingParam = new RenderingContext.RenderingParameter();

      var sampleDir = Runtime.IsRunningOnWindows ? @"C:\Users\hoge\Documents\" : "/home/hoge/";
      var execCommand = Runtime.IsRunningOnMono ? "mono BittenBaum.exe" : "BittenBaum.exe";

      var usageLines = new[] {
        VersionedApplicationName,
        ApplicationCopyright,
        ApplicationDistributionUrl,
        "",
        "使い方:",
        "  BittenBaum.exe ディレクトリ名 [オプション] [出力ファイル名]",
        "",
        "グラフのオプション:",
        "  次のオプションを組み合わせて使用できます。",
        "  /width:[幅]       出力する解析結果の画像の幅",
        string.Format("  /w:[幅]           デフォルト: {0}ピクセル", renderingParam.OutputWidth),
        "",
        "  /height:[高さ]    出力する解析結果の画像の高さ",
        string.Format("  /h:[高さ]         デフォルト: {0}ピクセル", renderingParam.OutputHeight),
        "",
        "  /depth:[深さ]     出力する解析結果の最大の深さ",
        "  /d:[深さ]         デフォルト: 指定なし(すべての階層を解析)",
        "",
        "  /dia:[内径]       出力するグラフの内径(0.1〜0.9)",
        string.Format("                    デフォルト: {0:F2}", renderingParam.InnerDiameterPercent),
        "",
        "  /colmode:[モード] 出力するグラフの色付けモード",
        string.Format("                    {0}: ディレクトリ別に色分けする", ColoringMode.Directory.ToString().ToLower()),
        string.Format("                    {0}: サイズの大きさに従って色分けする", ColoringMode.Volume.ToString().ToLower()),
        string.Format("                    デフォルト: {0}", renderingParam.ColoringMode.ToString().ToLower()),
        "",
        "  /colhue:[色相]    出力するグラフの色相(0〜359)",
        string.Format("                    /colmode:{0}の場合のみ有効", ColoringMode.Volume.ToString().ToLower()),
        string.Format("                    デフォルト: {0}", renderingParam.ColoringHue),
        "",
        "  /bg:[KnownColor]  指定したKnownColorを背景色とする",
        "                    有効な値はSystem.Drawing.KnownColorで定義",
        "                    されている値",
        "",
        "  /text:[KnownColor]指定したKnownColorを文字色とする",
        "                    有効な値はSystem.Drawing.KnownColorで定義",
        "                    されている値",
        "",
        "  /textsize:[サイズ]文字のサイズ(ポイント単位)",
        string.Format("                    デフォルト: {0}", renderingParam.FontSize),
        "",
        "  /driveusage       ルートディレクトリを解析する場合、ドライブ総容量に",
        "                    対するディレクトリの占有率でグラフ化する",
        "",
        "  /detail           微少なエントリも描画する",
        "",
        "  /nogradation      塗りつぶしのグラデーションを無効にする",
        "",
        "  /nolabel          グラフ内にエントリ名を描画しない",
        "",
        "出力形式に関するオプション:",
        "  /csv              CSV形式で結果を出力する",
        "",
        "  /xml              XML形式で結果を出力する",
        "",
        "  /individuate      ファイルとディレクトリを区別する",
        "  /i",
        "",
        "  /name             サイズ順ではなく名前順にソートする",
        "  /n",
        "",
        "  /fullpath         サイズ順ではなくフルパス名順にソートする",
        "  /f",
        "",
        "  /reverse          ソート順を逆にする",
        "",
        "動作に関するオプション:",
        "  /exclude:[Regex]  正規表現にマッチするディレクトリを除外して",
        "                    解析する",
        "",
#if false
        "  /fast             解析するディレクトリに含まれるエントリ数の",
        "                    計上を省略する",
        "",
#endif
        "  /silent           エラーを除く一切のメッセージを表示しない",
        "  /quiet",
        "  /s",
        "",
        "  /help, /?         使い方を表示する",
        "",
        "使用例:",
        string.Format("  {0} {1} analyzed.bmp", execCommand, sampleDir),
        string.Format("    {0}の解析結果をanalyzed.bmpに保存する", sampleDir),
        "",
        string.Format("  {0} C:\\ /driveusage analyzed.bmp", execCommand),
        "    Cドライブを解析し、各ディレクトリの占有率をグラフ化して",
        "    analyzed.bmpに保存する",
        "",
        string.Format("  {0} {1} /bg:black /text:white /dia:0.2 analyzed.png", execCommand, sampleDir),
        "    背景を黒、文字色を白、ディスクの中心を小さめにして、",
        "    PNG形式で保存する",
        "",
        string.Format("  {0} {1} /csv /i files.csv", execCommand, sampleDir),
        "    サイズが大きいファイルの一覧をCSV形式で出力する",
        "    ファイルとディレクトリは別々に分けた状態で出力する",
        "",
      };

      foreach (var line in usageLines) {
        Console.Error.WriteLine(line);
      }
    }
  }
}
