//  
//  Author:
//       smdn <smdn@smdn.jp>
// 
//  Copyright (c) 2003-2013 smdn
// 
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

using System;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Text;
using System.IO;

using Smdn.Formats.Ini;

using Smdn.Applications.BittenBaum.Analysis;
using Smdn.Applications.BittenBaum.Graphing;
using Smdn.Applications.BittenBaum.Rendering;
using Smdn.Applications.BittenBaum.UI;

namespace Smdn.Applications.BittenBaum {
  public static class Config {
    internal const string DefaultConfigFileName = "BittenBaum.ini";

    internal static readonly VisualizerGraphOptionDefaultsImpl VisualizerGraphOptionDefaults = new VisualizerGraphOptionDefaultsImpl();
    internal static readonly SaveGraphOptionDefaultsImpl SaveGraphOptionDefaults = new SaveGraphOptionDefaultsImpl();

    internal class VisualizerGraphOptionDefaultsImpl : GraphOptionDefaultsBase {
      protected override string SectionName {
        get { return "VisualizerGraphOptionDefaults"; }
      }

      private DirectoryEntryCompareMode entryCompareMode = DirectoryEntryCompareMode.BySize;
      private bool compareEntryType = true;
      private bool calculateRangeByDriveUsage = false;

      internal override void LoadFrom(IniDocument document)
      {
        base.LoadFrom(document);

        var section = document[SectionName];

        entryCompareMode            = section.Get("EntryCompareMode", entryCompareMode, EnumUtils.Parse<DirectoryEntryCompareMode>);
        compareEntryType            = section.Get("CompareEntryType", compareEntryType, Convert.ToBoolean);
        calculateRangeByDriveUsage  = section.Get("CalculateRangeByDriveUsage", calculateRangeByDriveUsage, Convert.ToBoolean);
      }

      internal override void SaveTo(IniDocument document)
      {
        base.SaveTo(document);

        var section = document[SectionName];

        section["EntryCompareMode"]           = entryCompareMode.ToString();
        section["CompareEntryType"]           = compareEntryType.ToString();
        section["CalculateRangeByDriveUsage"] = calculateRangeByDriveUsage.ToString();
      }

      internal override void Apply(GraphOptionDialog dialog)
      {
        base.Apply(dialog);

        dialog.EntryCompareMode           = entryCompareMode;
        dialog.CompareEntryType           = compareEntryType;
        dialog.CalculateRangeByDriveUsage = calculateRangeByDriveUsage;
      }

      internal bool Apply(GraphingContext.GraphingParameter param)
      {
        var newComparison = DirectoryEntryComparison.GetComparison(entryCompareMode,
                                                                   DirectoryEntryCompareOrder.Ascending,
                                                                   compareEntryType);

        var ret = (param.Comparison != newComparison ||
                   param.CalculateRangeByDriveUsage != calculateRangeByDriveUsage);

        param.Comparison = newComparison;
        param.CalculateRangeByDriveUsage = calculateRangeByDriveUsage;

        return ret;
      }

      internal override void Store(GraphOptionDialog dialog)
      {
        base.Store(dialog);

        entryCompareMode            = dialog.EntryCompareMode;
        compareEntryType            = dialog.CompareEntryType;
        calculateRangeByDriveUsage  = dialog.CalculateRangeByDriveUsage;
      }
    }

    internal class SaveGraphOptionDefaultsImpl : GraphOptionDefaultsBase {
      protected override string SectionName {
        get { return "SaveGraphOptionDefaults"; }
      }

      private int fontSize = RenderingContext.RenderingParameterDefaults.FontSize;
      private int outputSize = RenderingContext.RenderingParameterDefaults.OutputWidth;
      private bool drawNames = RenderingContext.RenderingParameterDefaults.DrawNames;
      private bool drawStatistics = RenderingContext.RenderingParameterDefaults.DrawStatistics;
      private float innerDiameterPercent = RenderingContext.RenderingParameterDefaults.InnerDiameterPercent;

      internal override void LoadFrom(IniDocument document)
      {
        base.LoadFrom(document);

        var section = document[SectionName];

        fontSize              = section.Get("FontSize", fontSize, Convert.ToInt32);
        outputSize            = section.Get("OutputSize", outputSize, Convert.ToInt32);
        drawNames             = section.Get("DrawNames", drawNames, Convert.ToBoolean);
        drawStatistics        = section.Get("DrawStatistics", drawStatistics, Convert.ToBoolean);
        innerDiameterPercent  = section.Get("InnerDiameterPercent", innerDiameterPercent, Convert.ToSingle);
      }

      internal override void SaveTo(IniDocument document)
      {
        base.SaveTo(document);

        var section = document[SectionName];

        section["FontSize"]               = fontSize.ToString();
        section["OutputSize"]             = outputSize.ToString();
        section["DrawNames"]              = drawNames.ToString();
        section["DrawStatistics"]         = drawStatistics.ToString();
        section["InnerDiameterPercent"]   = innerDiameterPercent.ToString();
      }

      internal override void Apply(GraphOptionDialog dialog)
      {
        base.Apply(dialog);

        dialog.FontSize               = fontSize;
        dialog.OutputSize             = outputSize;
        dialog.DrawNames              = drawNames;
        dialog.DrawStatistics         = drawStatistics;
        dialog.InnerDiameterPercent   = innerDiameterPercent;
      }

      internal override void Apply(RenderingContext.RenderingParameter param)
      {
        base.Apply(param);

        param.FontSize                = fontSize;
        param.OutputWidth             = outputSize;
        param.OutputHeight            = outputSize;
        param.DrawNames               = drawNames;
        param.DrawStatistics          = drawStatistics;
        param.InnerDiameterPercent    = innerDiameterPercent;
      }

      internal override void Store(GraphOptionDialog dialog)
      {
        base.Store(dialog);

        fontSize               = dialog.FontSize;
        outputSize             = dialog.OutputSize;
        drawNames              = dialog.DrawNames;
        drawStatistics         = dialog.DrawStatistics;
        innerDiameterPercent   = dialog.InnerDiameterPercent;
      }
    }

    internal abstract class GraphOptionDefaultsBase {
      protected abstract string SectionName { get; }

      private ColoringMode coloringMode = RenderingContext.RenderingParameterDefaults.ColoringMode;
      private int coloringHue = RenderingContext.RenderingParameterDefaults.ColoringHue;
      private bool drawHighQuality = true;
      private bool drawSmallEntries = RenderingContext.RenderingParameterDefaults.DrawSmallEntries;
      private Color backgroundColor = RenderingContext.RenderingParameterDefaults.BackgroundColor;
      private Color textColor = RenderingContext.RenderingParameterDefaults.TextColor;
      private DirectoryEntryCompareMode entryCompareMode = DirectoryEntryCompareMode.BySize;
      private bool compareEntryType = true;

      internal virtual void LoadFrom(IniDocument document)
      {
        var section = document[SectionName];

        coloringMode       = section.Get("ColoringMode", coloringMode, EnumUtils.Parse<ColoringMode>);
        coloringHue        = section.Get("ColoringHue", coloringHue, Convert.ToInt32);
        drawHighQuality    = section.Get("DrawHighQuality", drawHighQuality, Convert.ToBoolean);
        drawSmallEntries   = section.Get("DrawSmallEntries", drawSmallEntries, Convert.ToBoolean);
        backgroundColor    = section.Get("BackgroundColor", backgroundColor, Color.FromName);
        textColor          = section.Get("TextColor", textColor, Color.FromName);
        entryCompareMode   = section.Get("EntryCompareMode", entryCompareMode, EnumUtils.Parse<DirectoryEntryCompareMode>);
        compareEntryType   = section.Get("CompareEntryType", compareEntryType, Convert.ToBoolean);
      }

      internal virtual void SaveTo(IniDocument document)
      {
        var section = document[SectionName];

        section["ColoringMode"]       = coloringMode.ToString();
        section["ColoringHue"]        = coloringHue.ToString();
        section["DrawHighQuality"]    = drawHighQuality.ToString();
        section["DrawSmallEntries"]   = drawSmallEntries.ToString();
        section["BackgroundColor"]    = backgroundColor.Name;
        section["TextColor"]          = textColor.Name;
        section["EntryCompareMode"]   = entryCompareMode.ToString();
        section["CompareEntryType"]   = compareEntryType.ToString();
      }

      internal virtual void Apply(GraphOptionDialog dialog)
      {
        dialog.ColoringMode           = coloringMode;
        dialog.ColoringHue            = coloringHue;
        dialog.DrawHighQuality        = drawHighQuality;
        dialog.DrawSmallEntries       = drawSmallEntries;
        dialog.GraphBackgroundColor   = backgroundColor;
        dialog.EntryCompareMode       = entryCompareMode;
        dialog.CompareEntryType       = compareEntryType;
      }

      internal virtual void Apply(RenderingContext.RenderingParameter param)
      {
        param.ColoringMode = coloringMode;
        param.ColoringHue = coloringHue;

        if (drawHighQuality) {
          param.SmoothingMode     = SmoothingMode.AntiAlias;
          param.TextRenderingHint = TextRenderingHint.AntiAlias;
          param.InterpolationMode = InterpolationMode.HighQualityBicubic;
          param.EnableGradation   = true;
        }
        else {
          param.SmoothingMode     = SmoothingMode.HighSpeed;
          param.TextRenderingHint = TextRenderingHint.SystemDefault;
          param.InterpolationMode = InterpolationMode.NearestNeighbor;
          param.EnableGradation   = false;
        }

        param.DrawSmallEntries = drawSmallEntries;
        param.BackgroundColor = backgroundColor;
        param.TextColor = textColor;
      }

      internal virtual void Store(GraphOptionDialog dialog)
      {
        coloringMode           = dialog.ColoringMode;
        coloringHue            = dialog.ColoringHue;
        drawHighQuality        = dialog.DrawHighQuality;
        drawSmallEntries       = dialog.DrawSmallEntries;
        backgroundColor        = dialog.GraphBackgroundColor;
        textColor              = dialog.GraphTextColor;
        entryCompareMode       = dialog.EntryCompareMode;
        compareEntryType       = dialog.CompareEntryType;
      }
    }

    public static void Load(string searchDirectory)
    {
      configFilePath = Path.Combine(searchDirectory, DefaultConfigFileName);

      if (!File.Exists(configFilePath)) {
        Console.WriteLine("configuration file '{0}' not found", configFilePath);
        return;
      }

      try {
        var document = IniDocument.Load(configFilePath, StringComparer.Ordinal);

        VisualizerGraphOptionDefaults.LoadFrom(document);
        SaveGraphOptionDefaults.LoadFrom(document);
      }
      catch {
        // ignore any exceptions
      }
    }

    public static void Save()
    {
      try {
        var document = new IniDocument(StringComparer.Ordinal);

        document.DefaultSection["Version"] = BittenBaum.ApplicationVersion;

        VisualizerGraphOptionDefaults.SaveTo(document);
        SaveGraphOptionDefaults.SaveTo(document);

        document.Save(configFilePath);
      }
      catch {
        // ignore any exceptions
      }
    }

    private static string configFilePath;
  }
}

