//  
//  Author:
//       smdn <smdn@smdn.jp>
// 
//  Copyright (c) 2003-2013 smdn
// 
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

using System;
using System.Diagnostics;

using Smdn.Formats;
using Smdn.IO.FileSystems;

namespace Smdn.Applications.BittenBaum {
  using Analysis;
  using Formatting;
  using Graphing;
  using Rendering;

  public class Visualizer {
    private delegate ITaskContext PrepareTaskContext(bool resetParameter);
    private delegate void TaskStart(ITaskContext context);

    private TaskStart[] tasks;
    private PrepareTaskContext[] taskContextPreparer;

    private bool isCompleted = false;
    private int taskTotal = 0;
    private int taskCompleted = 0;
    private TaskContext.TaskStatus status = null;
    private readonly AnalysisContext analysisContext;
    private readonly GraphingContext graphingContext = new GraphingContext();
    private readonly RenderingContext renderingContext = new RenderingContext();
    private readonly FormattingContext formattingContext = new FormattingContext();

    public bool IsCompleted {
      get { return isCompleted; }
    }

    public TaskContext.TaskStatus Status {
      get { return status; }
    }

    public AnalysisContext AnalysisContext {
      get { return analysisContext; }
    }

    public GraphingContext GraphingContext {
      get { return graphingContext; }
    }

    public RenderingContext RenderingContext {
      get { return renderingContext; }
    }

    public FormattingContext FormattingContext {
      get { return formattingContext; }
    }

    public int TaskTotal {
      get { return taskTotal; }
    }

    public int TaskCompleted {
      get { return taskCompleted; }
    }

    public Visualizer(Directory directory)
      : this(new AnalysisContext(directory))
    {
    }

    private Visualizer(AnalysisContext analysisContext)
    {
      this.analysisContext = analysisContext;

      InitializeTasks();
    }

    private void InitializeTasks()
    {
      // タスク
      tasks = new TaskStart[] {
        delegate(ITaskContext context) {
          AnalysisContext analysisContext = context as AnalysisContext;

          Console.Write("Analyzing directory... ");

          // 解析
          var analyzer = new Analyzer(analysisContext);

          status = analysisContext.Status;

          analyzer.Analyze();

          status = null;

          var allEntryCount = analysisContext.Result.DirectoryTree.TotalEntryCount;

          ConsoleUtils.WriteProgress(allEntryCount, allEntryCount);
          Console.WriteLine();
          Console.WriteLine("Entries in this directory: {0}", allEntryCount);
          Console.WriteLine("Deepest directory depth: {0}", analysisContext.Result.DirectoryTree.MaxDepth);
        },

        delegate(ITaskContext context) {
          var graphingContext = context as GraphingContext;

          // サイズを計算(ツリー全体のエントリのサイズを一気に計算してキャッシュする)
          Console.Write("Calculating directory sizes(It might take a while)... ");

          context.Status.StatusText = "全体のサイズを計算中(少し時間がかかります)";

          status = graphingContext.Status;

          var totalFileSize = graphingContext.Parameter.DirectoryTree.GetTotalFileSize();

          Console.WriteLine();
          Console.WriteLine(string.Format(SIPrefixFormat.InvaliantInfo, "Directory size: {0:F} ({0:N0} bytes)", totalFileSize));

          Console.Write("Generating graph... ");

          // グラフを作成
          var grapher = new Grapher(graphingContext);

          grapher.Create();

          status = null;

          ConsoleUtils.WriteProgress((int)graphingContext.Status.Processed, (int)graphingContext.Status.Total);
          Console.WriteLine();
        },

        delegate(ITaskContext context) {
          var renderingContext = context as RenderingContext;

          Console.Write("Rendering graph... ");

          // グラフをレンダリング
          var renderer = new Renderer(renderingContext);

          status = renderingContext.Status;

          renderer.Render();

          status = null;

          ConsoleUtils.WriteProgress((int)renderingContext.Status.Processed, (int)renderingContext.Status.Total);
          Console.WriteLine();
        },

        delegate(ITaskContext context) {
          var formattingContext = context as FormattingContext;

          Console.Write("Formatting graph... ");

          // グラフを整形
          var formatter = new Formatter(formattingContext);

          status = formattingContext.Status;

          formatter.Format();

          status = null;

          ConsoleUtils.WriteProgress((int)formattingContext.Status.Processed, (int)formattingContext.Status.Total);
          Console.WriteLine();
        },
      };

      // 各タスクに引き渡すコンテキストを用意する処理
      taskContextPreparer = new PrepareTaskContext[] {
        delegate(bool reset) {
          return analysisContext;
        },

        delegate(bool reset) {
          if (reset)
            graphingContext.Parameter.DirectoryTree = analysisContext.Result.DirectoryTree;

          return graphingContext;
        },

        delegate(bool reset) {
          if (reset)
            renderingContext.Parameter.Graph = graphingContext.Result.Graph;

          return renderingContext;
        },

        delegate(bool reset) {
          if (reset) {
            formattingContext.Parameter.Graph = graphingContext.Result.Graph;
            formattingContext.Parameter.FormattingElements = graphingContext.Result.Graph.GetAllElements();
          }

          return formattingContext;
        },
      };
    }

    /// <summary>
    /// ディレクトリの解析、グラフの作成、グラフの描画を行う
    /// </summary>
    public void Visualize()
    {
      ExecuteTask(new[] {0, 1, 2}, true);
    }

    /// <summary>
    /// ディレクトリの解析、グラフの作成、グラフの整形を行う
    /// </summary>
    public void Format()
    {
      ExecuteTask(new[] {0, 1, 3}, true);
    }

    /// <summary>
    /// グラフの作成を行う
    /// </summary>
    public void Graph(bool resetParameter)
    {
      ExecuteTask(new[] {1}, resetParameter);
    }

    /// <summary>
    /// グラフの描画を行う
    /// </summary>
    public void Render(bool resetParameter)
    {
      ExecuteTask(new[] {2}, resetParameter);
    }

    /// <summary>
    /// グラフの整形を行う
    /// </summary>
    public void Format(bool resetParameter)
    {
      ExecuteTask(new[] {3}, resetParameter);
    }

    /// <summary>
    /// 指定されたインデックスのタスクを実行する
    /// </summary>
    private void ExecuteTask(int[] executeTaskIndices, bool resetParameter)
    {
      isCompleted = false;
      taskTotal = executeTaskIndices.Length;
      taskCompleted = 0;

      // 各タスクを実行
      for (var task = 0; task < executeTaskIndices.Length; task++) {
        var stopwatch = Stopwatch.StartNew();

        try {
          var taskIndex = executeTaskIndices[task];

          // タスクに引き渡すコンテキストを準備
          ITaskContext context = taskContextPreparer[taskIndex](resetParameter);

          // タスクを起動
          tasks[taskIndex](context);

          taskCompleted++;
        }
        catch (System.Threading.ThreadAbortException) {
          Console.Error.WriteLine("Abort requested");

          throw;
        }
        catch (Exception ex) {
          Console.Error.WriteLine(ex);

          throw ex;
        }

        // 経過時間を出力
        stopwatch.Stop();

        Console.WriteLine("  Elapsed: {0}", stopwatch.Elapsed);
        Console.WriteLine();

        stopwatch.Reset();
      }

      // 完了
      isCompleted = true;

      return;
    }
  }
}
